import asyncio
import gzip
import json
import logging
import os
import sys
from collections.abc import Callable, Generator, Sequence
from contextlib import ExitStack
from functools import partial
from pathlib import Path
from types import ModuleType
from typing import Any, TypeVar
from unittest.mock import MagicMock, patch

import bluesky.plan_stubs as bps
import numpy
import pydantic
import pytest
from bluesky.simulators import RunEngineSimulator
from bluesky.utils import Msg, MsgGenerator
from dodal.beamlines import aithre, i03
from dodal.common.beamlines import beamline_utils
from dodal.common.beamlines.beamline_utils import clear_devices
from dodal.common.beamlines.commissioning_mode import set_commissioning_signal
from dodal.common.beamlines.config_client import get_config_client
from dodal.devices.aperturescatterguard import (
    AperturePosition,
    ApertureScatterguard,
    ApertureValue,
)
from dodal.devices.attenuator.attenuator import (
    BinaryFilterAttenuator,
    EnumFilterAttenuator,
)
from dodal.devices.attenuator.filter_selections import (
    I24FilterOneSelections,
    I24FilterTwoSelections,
)
from dodal.devices.backlight import Backlight
from dodal.devices.baton import Baton
from dodal.devices.beamlines.i03 import Beamstop, BeamstopPositions
from dodal.devices.beamlines.i03.beamsize import Beamsize
from dodal.devices.beamlines.i03.dcm import DCM
from dodal.devices.beamlines.i04.transfocator import Transfocator
from dodal.devices.beamsize.beamsize import BeamsizeBase
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import FastGridScanCommon
from dodal.devices.flux import Flux
from dodal.devices.oav.oav_detector import OAV, OAVConfigBeamCentre
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot, SampleLocation
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.scintillator import Scintillator
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.thawer import Thawer
from dodal.devices.undulator import UndulatorInKeV
from dodal.devices.webcam import Webcam
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from dodal.devices.zocalo.zocalo_results import _NO_SAMPLE_ID
from dodal.log import LOGGER as DODAL_LOGGER
from dodal.log import set_up_all_logging_handlers
from dodal.testing.fixtures.config_server import fake_config_server_get_file_contents
from dodal.utils import AnyDeviceFactory, collect_factories
from event_model.documents import Event, EventDescriptor, RunStart, RunStop
from ophyd_async.core import (
    AsyncStatus,
    Device,
    DeviceVector,
    Reference,
    completed_status,
    get_mock_put,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.epics.motor import Motor
from ophyd_async.fastcs.panda import DatasetTable, PandaHdf5DatasetType
from PIL import Image
from pydantic.dataclasses import dataclass
from scanspec.core import Path as ScanPath
from scanspec.specs import Line

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanPlane,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    EnvironmentConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.parameters.rotation import RotationScan
from mx_bluesky.common.utils.exceptions import CrystalNotFoundError
from mx_bluesky.common.utils.log import (
    ALL_LOGGERS,
    ISPYB_ZOCALO_CALLBACK_LOGGER,
    LOGGER,
    NEXUS_LOGGER,
    _get_logging_dirs,
    do_default_logging_setup,
)
from mx_bluesky.hyperion.baton_handler import HYPERION_USER
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

pytest_plugins = ["tests.expeye_helpers"]

i03.DAQ_CONFIGURATION_PATH = "tests/test_data/test_daq_configuration"

TEST_GRAYLOG_PORT = 5555
TEST_VISIT = "cm1234-67"
TEST_CONTAINER = 8
TEST_SESSION_ID = 90
EXPECTED_START_TIME = "2024-02-08 14:03:59"
EXPECTED_END_TIME = "2024-02-08 14:04:01"
TEST_SAMPLE_ID = 364758
TEST_BARCODE = "12345A"

TEST_RESULT_LARGE = [
    {
        "centre_of_mass": [1, 2, 3],
        "max_voxel": [1, 2, 3],
        "max_count": 105062,
        "n_voxels": 35,
        "total_count": 2387574,
        "bounding_box": [[2, 2, 2], [8, 8, 7]],
        "sample_id": _NO_SAMPLE_ID,
    }
]
TEST_RESULT_MEDIUM = [
    {
        "centre_of_mass": [1, 2, 3],
        "max_voxel": [2, 4, 5],
        "max_count": 50000,
        "n_voxels": 35,
        "total_count": 100000,
        "bounding_box": [[1, 2, 3], [3, 4, 4]],
        "sample_id": _NO_SAMPLE_ID,
    }
]
TEST_RESULT_SMALL = [
    {
        "centre_of_mass": [1, 2, 3],
        "max_voxel": [1, 2, 3],
        "max_count": 1000,
        "n_voxels": 35,
        "total_count": 1000,
        "bounding_box": [[2, 2, 2], [3, 3, 3]],
        "sample_id": _NO_SAMPLE_ID,
    }
]
TEST_RESULT_BELOW_THRESHOLD = [
    {
        "centre_of_mass": [2, 3, 4],
        "max_voxel": [2, 3, 4],
        "max_count": 2,
        "n_voxels": 1,
        "total_count": 2,
        "bounding_box": [[1, 2, 3], [2, 3, 4]],
        "sample_id": _NO_SAMPLE_ID,
    }
]

# These are the uncorrected coordinate from zocalo
TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX = [
    {
        "centre_of_mass": [0.5, 0.5, 0.5],
        "max_voxel": [0, 0, 0],
        "max_count": 50000,
        "n_voxels": 35,
        "total_count": 100000,
        "bounding_box": [[0, 0, 0], [3, 4, 4]],
        "sample_id": _NO_SAMPLE_ID,
    }
]
# These are the uncorrected coordinate from zocalo
TEST_RESULT_IN_BOUNDS_TOP_LEFT_GRID_CORNER = [
    {
        "centre_of_mass": [0.0, 0.0, 0.0],
        "max_voxel": [0, 0, 0],
        "max_count": 50000,
        "n_voxels": 35,
        "total_count": 100000,
        "bounding_box": [[0, 0, 0], [3, 4, 4]],
        "sample_id": _NO_SAMPLE_ID,
    }
]
# These are the uncorrected coordinate from zocalo
TEST_RESULT_OUT_OF_BOUNDS_COM = [
    {
        "centre_of_mass": [-0.0001, -0.0001, -0.0001],
        "max_voxel": [0, 0, 0],
        "max_count": 50000,
        "n_voxels": 35,
        "total_count": 100000,
        "bounding_box": [[0, 0, 0], [3, 4, 4]],
        "sample_id": _NO_SAMPLE_ID,
    }
]
# These are the uncorrected coordinate from zocalo
TEST_RESULT_OUT_OF_BOUNDS_BB = [
    {
        "centre_of_mass": [0, 0, 0],
        "max_voxel": [0, 0, 0],
        "max_count": 50000,
        "n_voxels": 35,
        "total_count": 100000,
        "bounding_box": [[-1, -1, -1], [3, 4, 4]],
        "sample_id": _NO_SAMPLE_ID,
    }
]

MOCK_DAQ_CONFIG_PATH = "tests/test_data/test_daq_configuration"
mock_paths = [
    ("DAQ_CONFIGURATION_PATH", MOCK_DAQ_CONFIG_PATH),
    ("ZOOM_PARAMS_FILE", "tests/test_data/test_jCameraManZoomLevels.xml"),
    ("DISPLAY_CONFIG", f"{MOCK_DAQ_CONFIG_PATH}/display.configuration"),
]


@dataclass(frozen=True)
class SimConstants:
    BEAMLINE = "BL03S"
    # The following are values present in the system test ispyb database
    ST_VISIT = "cm14451-2"
    ST_SAMPLE_ID = 398810
    ST_MSP_SAMPLE_IDS = [398816, 398819]
    ST_CONTAINER_ID = 34864


@pytest.fixture(autouse=True, scope="session")
def ispyb_config_path():
    ispyb_config_path = os.environ.get(
        "ISPYB_CONFIG_PATH", "tests/test_data/test_config.cfg"
    )
    with patch.dict(os.environ, {"ISPYB_CONFIG_PATH": ispyb_config_path}):
        yield ispyb_config_path


@pytest.fixture(scope="session")
def active_device_factories() -> set[AnyDeviceFactory]:
    """Obtain the set of device factories that should have their caches cleared
    after every test invocation.
    Override this in sub-packages for the specific beamlines under test."""
    return device_factories_for_beamline(i03)


def device_factories_for_beamline(beamline_module: ModuleType) -> set[AnyDeviceFactory]:
    return {
        f
        for f in collect_factories(beamline_module, include_skipped=True).values()
        if hasattr(f, "cache_clear")
    }


@pytest.fixture(scope="function", autouse=True)
def clear_device_factory_caches_after_every_test(active_device_factories):
    yield None
    for f in active_device_factories:
        f.cache_clear()  # type: ignore


def replace_all_tmp_paths(d: dict[str, Any], tmp_path: Path):
    d = d.copy()
    for k, v in d.items():
        if isinstance(v, str):
            d[k] = v.replace("{tmp_data}", str(tmp_path))
        elif isinstance(v, dict):
            d[k] = replace_all_tmp_paths(v, tmp_path)
    return d


def raw_params_from_file(filename, tmp_path):
    with open(filename) as f:
        loads = replace_all_tmp_paths(json.loads(f.read()), tmp_path)
        return loads


def create_dummy_scan_spec():
    x_line = Line("sam_x", 0, 10, 10)
    y_line = Line("sam_y", 10, 20, 20)
    z_line = Line("sam_z", 30, 50, 30)

    specs = [y_line * ~x_line, z_line * ~x_line]
    specs = [ScanPath(spec.calculate()) for spec in specs]
    return [spec.consume().midpoints for spec in specs]


def _reset_loggers(loggers):
    """Clear all handlers and tear down the logging hierarchy, leave logger references intact."""
    clear_log_handlers(loggers)
    for logger in loggers:
        if logger.name != "Hyperion" and logger.name != "MX-Bluesky":
            # Hyperion parent is configured on module import, do not remove
            logger.parent = logging.getLogger()


def clear_log_handlers(loggers: Sequence[logging.Logger]):
    for logger in loggers:
        for handler in logger.handlers:
            handler.close()
        logger.handlers.clear()


def pytest_runtest_setup(item):
    markers = [m.name for m in item.own_markers]
    if item.config.getoption("logging") and "skip_log_setup" not in markers:
        if LOGGER.handlers == []:
            if DODAL_LOGGER.handlers == []:
                print("Initialising Hyperion logger for tests")
                do_default_logging_setup("dev_log.py", TEST_GRAYLOG_PORT, dev_mode=True)
        logging_path, _ = _get_logging_dirs(True)
        if ISPYB_ZOCALO_CALLBACK_LOGGER.handlers == []:
            print("Initialising ISPyB logger for tests")
            set_up_all_logging_handlers(
                ISPYB_ZOCALO_CALLBACK_LOGGER,
                logging_path,
                "hyperion_ispyb_callback.log",
                True,
                10000,
            )
        if NEXUS_LOGGER.handlers == []:
            print("Initialising nexus logger for tests")
            set_up_all_logging_handlers(
                NEXUS_LOGGER,
                logging_path,
                "hyperion_ispyb_callback.log",
                True,
                10000,
            )
    else:
        print("Skipping log setup for log test - deleting existing handlers")
        _reset_loggers([*ALL_LOGGERS, DODAL_LOGGER])
        handler = logging.StreamHandler()
        handler.setLevel(logging.INFO)
        handler.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)s %(name)s: %(message)s")
        )
        DODAL_LOGGER.addHandler(handler)


def pytest_runtest_teardown(item):
    if "dodal.common.beamlines.beamline_utils" in sys.modules:
        sys.modules["dodal.common.beamlines.beamline_utils"].clear_devices()
    markers = [m.name for m in item.own_markers]
    if "skip_log_setup" in markers:
        _reset_loggers([*ALL_LOGGERS, DODAL_LOGGER])


def pass_on_mock(motor: Motor, call_log: MagicMock | None = None):
    def _pass_on_mock(value: float, wait: bool):
        set_mock_value(motor.user_readback, value)
        if call_log is not None:
            call_log(value, wait=wait)

    return _pass_on_mock


@pytest.fixture
def beamline_parameters():
    return fake_config_server_get_file_contents(
        "tests/test_data/test_beamline_parameters.txt", dict
    )


@pytest.fixture(autouse=True)
def i03_beamline_parameters():
    """Fix default i03 beamline parameters to refer to a test file not the /dls_sw folder"""
    with patch.dict(
        "dodal.common.beamlines.beamline_parameters.BEAMLINE_PARAMETER_PATHS",
        {"i03": "tests/test_data/test_beamline_parameters.txt"},
    ) as params:
        with ExitStack() as context_stack:
            for context_mgr in [
                patch(f"dodal.beamlines.i03.{name}", value, create=True)
                for name, value in mock_paths
            ]:
                context_stack.enter_context(context_mgr)
            yield params


@pytest.fixture(autouse=True)
def test_beamline_parameters():
    """Fix default test beamline parameters to refer to a test file not the /dls_sw folder"""
    with patch.dict(
        "dodal.common.beamlines.beamline_parameters.BEAMLINE_PARAMETER_PATHS",
        {"test": "tests/test_data/test_beamline_parameters.txt"},
    ) as params:
        yield params


@pytest.fixture
def hyperion_fgs_params(tmp_path):
    return HyperionSpecifiedThreeDGridScan(
        **(
            raw_params_from_file(
                "tests/test_data/parameter_json_files/good_test_parameters.json",
                tmp_path,
            )
        )
    )


@pytest.fixture
def eiger():
    eiger = i03.eiger.build(mock=True)
    eiger.stage = MagicMock(side_effect=lambda: completed_status())
    eiger.do_arm.set = MagicMock(side_effect=lambda _: completed_status())
    eiger.unstage = MagicMock(side_effect=lambda: completed_status())
    return eiger


@pytest.fixture
def smargon() -> Generator[Smargon, None, None]:
    smargon = i03.smargon.build(connect_immediately=True, mock=True)
    # Initial positions, needed for stub_offsets
    set_mock_value(smargon.stub_offsets.center_at_current_position.disp, 0)
    yield smargon
    clear_devices()


@pytest.fixture
def aithre_gonio():
    return aithre.goniometer.build(connect_immediately=True, mock=True)


@pytest.fixture
def zebra():
    return i03.zebra.build(connect_immediately=True, mock=True)


@pytest.fixture
def zebra_shutter():
    shutter = i03.sample_shutter.build(connect_immediately=True, mock=True)

    def put_sample_shutter(value, **kwargs):
        set_mock_value(shutter.position_readback, value)

    get_mock_put(shutter._manual_position_setpoint).side_effect = put_sample_shutter
    return shutter


@pytest.fixture
def backlight():
    backlight = i03.backlight.build(connect_immediately=True, mock=True)
    backlight.TIME_TO_MOVE_S = 0.001
    return backlight


@pytest.fixture
def baton():
    baton = i03.baton.build(connect_immediately=True, mock=True)
    set_mock_value(baton.requested_user, HYPERION_USER)
    set_mock_value(baton.current_user, HYPERION_USER)
    return baton


@pytest.fixture
def baton_in_commissioning_mode(baton: Baton):
    set_commissioning_signal(baton.commissioning)
    set_mock_value(baton.commissioning, True)
    yield baton
    set_commissioning_signal(None)


@pytest.fixture
def fast_grid_scan():
    scan = i03.zebra_fast_grid_scan.build(mock=True, connect_immediately=True)
    for signal in [scan.x_scan_valid, scan.y_scan_valid, scan.z_scan_valid]:
        set_mock_value(signal, 1)
    return scan


@pytest.fixture
def detector_motion():
    return i03.detector_motion.build(mock=True, connect_immediately=True)


@pytest.fixture
def undulator():
    undulator = i03.undulator.build(mock=True, connect_immediately=True)
    return undulator


@pytest.fixture
def s4_slit_gaps():
    return i03.s4_slit_gaps.build(mock=True, connect_immediately=True)


@pytest.fixture
def synchrotron():
    synchrotron = i03.synchrotron.build(mock=True, connect_immediately=True)
    set_mock_value(synchrotron.synchrotron_mode, SynchrotronMode.USER)
    set_mock_value(synchrotron.top_up_start_countdown, 10)
    return synchrotron


@pytest.fixture
def oav(test_config_files):
    parameters = OAVConfigBeamCentre(
        test_config_files["zoom_params_file"], test_config_files["display_config"]
    )
    oav = i03.oav.build(mock=True, connect_immediately=True, params=parameters)

    set_mock_value(oav.zoom_controller.level, "5.0x")
    set_mock_value(oav.grid_snapshot.x_size, 1024)
    set_mock_value(oav.grid_snapshot.y_size, 768)

    oav.snapshot.trigger = MagicMock(side_effect=lambda: completed_status())
    oav.grid_snapshot.trigger = MagicMock(side_effect=lambda: completed_status())
    yield oav


@pytest.fixture
def flux():
    return i03.flux.build(connect_immediately=True, mock=True)


@pytest.fixture
def pin_tip():
    return i03.pin_tip_detection.build(connect_immediately=True, mock=True)


@pytest.fixture
def ophyd_pin_tip_detection():
    return i03.pin_tip_detection.build(connect_immediately=True, mock=True)


@pytest.fixture()
def transfocator():
    with init_devices(mock=True):
        transfocator = Transfocator("", "")
    transfocator.set = MagicMock(side_effect=lambda _: completed_status())
    return transfocator


@pytest.fixture
def robot():
    robot = i03.robot.build(connect_immediately=True, mock=True)
    set_mock_value(robot.barcode, "BARCODE")

    @AsyncStatus.wrap
    async def fake_load(val: SampleLocation):
        if val is not None:  # type: ignore
            set_mock_value(robot.current_pin, val.pin)
            set_mock_value(robot.current_puck, val.puck)
            set_mock_value(robot.sample_id, await robot.next_sample_id.get_value())

    robot.set = MagicMock(side_effect=fake_load)
    return robot


@pytest.fixture
def scintillator(aperture_scatterguard):
    with init_devices(mock=True):
        scintillator = Scintillator(
            "",
            aperture_scatterguard=Reference(aperture_scatterguard),
            beamline_parameters=MagicMock(),
            name="scintillator",
        )
    return scintillator


@pytest.fixture
def attenuator():
    attenuator = i03.attenuator.build(connect_immediately=True, mock=True)
    set_mock_value(attenuator.actual_transmission, 0.49118047952)

    @AsyncStatus.wrap
    async def fake_attenuator_set(val):
        set_mock_value(attenuator.actual_transmission, val)

    attenuator.set = MagicMock(side_effect=fake_attenuator_set)

    yield attenuator


@pytest.fixture
def beamstop_phase1(
    beamline_parameters: dict[str, Any],
    sim_run_engine: RunEngineSimulator,
) -> Generator[Beamstop, Any, Any]:
    with patch(
        "dodal.beamlines.i03.get_beamline_parameters",
        return_value=beamline_parameters,
    ):
        beamstop = i03.beamstop.build(connect_immediately=True, mock=True)

        set_mock_value(beamstop.x_mm.user_readback, 1.52)
        set_mock_value(beamstop.y_mm.user_readback, 44.78)
        set_mock_value(beamstop.z_mm.user_readback, 30.0)

        # sim_run_engine.add_read_handler_for(
        #     beamstop.selected_pos, BeamstopPositions.DATA_COLLECTION
        # )
        # Can uncomment and remove below when https://github.com/bluesky/bluesky/issues/1906 is fixed
        def locate_beamstop(_):
            return {"readback": BeamstopPositions.DATA_COLLECTION}

        sim_run_engine.add_handler(
            "locate", locate_beamstop, beamstop.selected_pos.name
        )

        yield beamstop
        beamline_utils.clear_devices()


@pytest.fixture
def xbpm_feedback(
    baton: Baton,  # Ensure baton is cached with mock configuration
):
    xbpm = i03.xbpm_feedback.build(connect_immediately=True, mock=True)
    xbpm.trigger = MagicMock(side_effect=lambda: completed_status())
    yield xbpm
    beamline_utils.clear_devices()


def set_up_dcm(dcm: DCM, sim_run_engine: RunEngineSimulator):
    set_mock_value(dcm.energy_in_keV.user_readback, 12.7)
    set_mock_value(dcm.xtal_1.pitch_in_mrad.user_readback, 1)
    set_mock_value(dcm.crystal_metadata_d_spacing_a, 3.13475)
    sim_run_engine.add_read_handler_for(dcm.crystal_metadata_d_spacing_a, 3.13475)
    return dcm


@pytest.fixture
def dcm(sim_run_engine):
    dcm = i03.dcm.build(connect_immediately=True, mock=True)
    set_up_dcm(dcm, sim_run_engine)
    yield dcm


@pytest.fixture
def vfm():
    vfm = i03.vfm.build(connect_immediately=True, mock=True)
    vfm.bragg_to_lat_lookup_table_path = (
        "tests/test_data/test_beamline_vfm_lat_converter.txt"
    )
    return vfm


@pytest.fixture
def lower_gonio(
    sim_run_engine: RunEngineSimulator,
):
    lower_gonio = i03.lower_gonio.build(connect_immediately=True, mock=True)

    # Replace when https://github.com/bluesky/bluesky/issues/1906 is fixed
    def locate_gonio(_):
        return {"readback": 0}

    sim_run_engine.add_handler("locate", locate_gonio, lower_gonio.x.name)
    sim_run_engine.add_handler("locate", locate_gonio, lower_gonio.y.name)
    sim_run_engine.add_handler("locate", locate_gonio, lower_gonio.z.name)
    return lower_gonio


@pytest.fixture
def mirror_voltages():
    voltages = i03.mirror_voltages.build(connect_immediately=True, mock=True)
    voltages.voltage_lookup_table_path = "tests/test_data/test_mirror_focus.json"
    for vc in voltages.vertical_voltages.values():
        vc.set = MagicMock(side_effect=lambda _: completed_status())
    for vc in voltages.horizontal_voltages.values():
        vc.set = MagicMock(side_effect=lambda _: completed_status())
    yield voltages
    beamline_utils.clear_devices()


@pytest.fixture
def undulator_dcm(sim_run_engine, dcm, undulator):
    undulator_dcm = i03.undulator_dcm.build(
        connect_immediately=True,
        mock=True,
        daq_configuration_path="tests/test_data/test_daq_configuration",
        dcm=dcm,
        undulator=undulator,
    )
    set_up_dcm(undulator_dcm.dcm_ref(), sim_run_engine)
    yield undulator_dcm
    # beamline_utils.clear_devices()


@pytest.fixture
def webcam() -> Generator[Webcam, Any, Any]:
    webcam = i03.webcam.build(connect_immediately=True, mock=True)
    with patch.object(webcam, "_get_and_write_image"):
        yield webcam


@pytest.fixture
def thawer() -> Generator[Thawer, Any, Any]:
    yield i03.thawer.build(connect_immediately=True, mock=True)


@pytest.fixture
def sample_shutter() -> Generator[ZebraShutter, Any, Any]:
    yield i03.sample_shutter.build(connect_immediately=True, mock=True)


@pytest.fixture
async def aperture_scatterguard():
    positions = {
        ApertureValue.LARGE: AperturePosition(
            aperture_x=0,
            aperture_y=1,
            aperture_z=2,
            scatterguard_x=3,
            scatterguard_y=4,
            diameter=100,
        ),
        ApertureValue.MEDIUM: AperturePosition(
            aperture_x=5,
            aperture_y=6,
            aperture_z=2,
            scatterguard_x=8,
            scatterguard_y=9,
            diameter=50,
        ),
        ApertureValue.SMALL: AperturePosition(
            aperture_x=10,
            aperture_y=11,
            aperture_z=2,
            scatterguard_x=13,
            scatterguard_y=14,
            diameter=20,
        ),
        ApertureValue.OUT_OF_BEAM: AperturePosition(
            aperture_x=15,
            aperture_y=16,
            aperture_z=2,
            scatterguard_x=18,
            scatterguard_y=19,
            diameter=0,
        ),
        ApertureValue.PARKED: AperturePosition(
            aperture_x=20,
            aperture_y=25,
            aperture_z=0,
            scatterguard_x=36,
            scatterguard_y=56,
            diameter=0,
        ),
    }
    with (
        patch(
            "dodal.beamlines.i03.load_positions_from_beamline_parameters",
            return_value=positions,
        ),
        patch(
            "dodal.beamlines.i03.AperturePosition.tolerances_from_gda_params",
            return_value=AperturePosition(
                aperture_x=0.1,
                aperture_y=0.1,
                aperture_z=0.1,
                scatterguard_x=0.1,
                scatterguard_y=0.1,
            ),
        ),
    ):
        ap_sg = i03.aperture_scatterguard.build(connect_immediately=True, mock=True)

    await ap_sg.aperture.z.set(2)
    await ap_sg.selected_aperture.set(ApertureValue.SMALL)

    set_mock_value(ap_sg.aperture.small, 1)
    return ap_sg


@pytest.fixture()
async def beamsize(aperture_scatterguard: ApertureScatterguard):
    return Beamsize(aperture_scatterguard, name="beamsize")


@pytest.fixture()
def test_config_files():
    return {
        "zoom_params_file": "tests/test_data/test_jCameraManZoomLevels.xml",
        "oav_config_json": "tests/test_data/test_OAVCentring.json",
        "display_config": "tests/test_data/test_display.configuration",
    }


@pytest.fixture()
def fake_create_devices(
    beamstop_phase1: Beamstop,
    eiger: EigerDetector,
    smargon: Smargon,
    zebra: Zebra,
    detector_motion: DetectorMotion,
    aperture_scatterguard: ApertureScatterguard,
    backlight: Backlight,
):
    mock_omega_sets = MagicMock(side_effect=lambda _: completed_status())

    smargon.omega.velocity.set = mock_omega_sets
    smargon.omega.set = mock_omega_sets

    devices = {
        "beamstop": beamstop_phase1,
        "eiger": eiger,
        "gonio": smargon,
        "zebra": zebra,
        "detector_motion": detector_motion,
        "backlight": backlight,
        "ap_sg": aperture_scatterguard,
    }
    return devices


@pytest.fixture()
def fake_create_rotation_devices(
    beamstop_phase1: Beamstop,
    eiger: EigerDetector,
    smargon: Smargon,
    zebra: Zebra,
    detector_motion: DetectorMotion,
    backlight: Backlight,
    attenuator: BinaryFilterAttenuator,
    flux: Flux,
    undulator: UndulatorInKeV,
    aperture_scatterguard: ApertureScatterguard,
    synchrotron: Synchrotron,
    s4_slit_gaps: S4SlitGaps,
    dcm: DCM,
    robot: BartRobot,
    oav: OAV,
    sample_shutter: ZebraShutter,
    xbpm_feedback: XBPMFeedback,
    thawer: Thawer,
    beamsize: BeamsizeBase,
):
    set_mock_value(smargon.omega.max_velocity, 131)
    undulator.set = MagicMock(side_effect=lambda _: completed_status())
    return RotationScanComposite(
        attenuator=attenuator,
        backlight=backlight,
        beamsize=beamsize,
        beamstop=beamstop_phase1,
        dcm=dcm,
        detector_motion=detector_motion,
        eiger=eiger,
        flux=flux,
        gonio=smargon,
        undulator=undulator,
        aperture_scatterguard=aperture_scatterguard,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        zebra=zebra,
        robot=robot,
        oav=oav,
        sample_shutter=sample_shutter,
        xbpm_feedback=xbpm_feedback,
        thawer=thawer,
    )


@pytest.fixture
def zocalo():
    zoc = i03.zocalo.build(connect_immediately=True, mock=True)
    zoc.stage = MagicMock(side_effect=lambda: completed_status())
    zoc.unstage = MagicMock(side_effect=lambda: completed_status())
    return zoc


@pytest.fixture
async def enum_attenuator() -> EnumFilterAttenuator:
    with init_devices(mock=True):
        attenuator = EnumFilterAttenuator(
            "", filter_selection=(I24FilterOneSelections, I24FilterTwoSelections)
        )

    @AsyncStatus.wrap
    async def fake_attenuator_set(val):
        set_mock_value(attenuator.actual_transmission, val)

    attenuator.set = MagicMock(side_effect=fake_attenuator_set)
    return attenuator


@pytest.fixture
async def panda():
    class MockBlock(Device):
        def __init__(
            self,
            prefix: str,
            name: str = "",
            attributes: dict[str, Any] = {},  # noqa
        ):
            for name, dtype in attributes.items():
                setattr(self, name, epics_signal_rw(dtype, "", ""))
            super().__init__(name)

    def mock_vector_block(n, attributes):
        return DeviceVector(
            {i: MockBlock(f"{i}", f"{i}", attributes) for i in range(n)}
        )

    async def set_mock_blocks(
        panda, mock_blocks: dict[str, tuple[int, dict[str, Any]]]
    ):
        for name, block in mock_blocks.items():
            n, attrs = block
            block = mock_vector_block(n, attrs)
            await block.connect(mock=True)
            setattr(panda, name, block)

    async def create_mock_signals(devices_and_signals: dict[Device, dict[str, Any]]):
        for device, signals in devices_and_signals.items():
            for name, dtype in signals.items():
                sig = epics_signal_rw(dtype, name, name)
                await sig.connect(mock=True)
                setattr(device, name, sig)

    panda = i03.panda.build(connect_immediately=True, mock=True)
    await set_mock_blocks(
        panda,
        {
            "inenc": (8, {"setp": float}),
            "clock": (8, {"period": float}),
            "counter": (8, {"enable": str}),
        },
    )
    await create_mock_signals(
        {
            panda.pcap: {"enable": str},
            **{panda.pulse[i]: {"enable": str} for i in panda.pulse.keys()},
        }
    )

    set_mock_value(
        panda.data.datasets,
        DatasetTable(name=["name"], dtype=[PandaHdf5DatasetType.FLOAT_64]),
    )

    return panda


@pytest.fixture
def oav_parameters_for_rotation(test_config_files) -> OAVParameters:
    return OAVParameters(oav_config_json=test_config_files["oav_config_json"])


async def async_status_done():
    await asyncio.sleep(0)


def mock_gridscan_kickoff_complete(gridscan: FastGridScanCommon):
    gridscan.kickoff = MagicMock(return_value=async_status_done)
    gridscan.complete = MagicMock(return_value=async_status_done)


@pytest.fixture
def panda_fast_grid_scan():
    scan = i03.panda_fast_grid_scan.build(connect_immediately=True, mock=True)
    for signal in [scan.x_scan_valid, scan.y_scan_valid, scan.z_scan_valid]:
        set_mock_value(signal, 1)
    return scan


@pytest.fixture
async def hyperion_flyscan_xrc_composite(
    smargon: Smargon,
    hyperion_fgs_params: HyperionSpecifiedThreeDGridScan,
    attenuator,
    xbpm_feedback,
    synchrotron,
    aperture_scatterguard,
    zocalo,
    dcm,
    panda,
    backlight,
    s4_slit_gaps,
    fast_grid_scan,
    panda_fast_grid_scan,
    beamsize,
) -> HyperionFlyScanXRayCentreComposite:
    fake_composite = HyperionFlyScanXRayCentreComposite(
        aperture_scatterguard=aperture_scatterguard,
        attenuator=attenuator,
        backlight=backlight,
        dcm=dcm,
        # We don't use the eiger fixture here because .unstage() is used in some tests
        eiger=i03.eiger.build(mock=True),
        zebra_fast_grid_scan=fast_grid_scan,
        flux=i03.flux.build(connect_immediately=True, mock=True),
        s4_slit_gaps=s4_slit_gaps,
        gonio=smargon,
        undulator=i03.undulator.build(connect_immediately=True, mock=True),
        synchrotron=synchrotron,
        xbpm_feedback=xbpm_feedback,
        zebra=i03.zebra.build(connect_immediately=True, mock=True),
        zocalo=zocalo,
        panda=panda,
        panda_fast_grid_scan=panda_fast_grid_scan,
        robot=i03.robot.build(connect_immediately=True, mock=True),
        sample_shutter=i03.sample_shutter.build(connect_immediately=True, mock=True),
        beamsize=beamsize,
    )

    fake_composite.eiger.stage = MagicMock(side_effect=lambda: completed_status())
    # unstage should be mocked on a per-test basis because several rely on unstage
    fake_composite.eiger.set_detector_parameters(hyperion_fgs_params.detector_params)
    fake_composite.eiger.stop_odin_when_all_frames_collected = MagicMock()
    fake_composite.eiger.odin.check_and_wait_for_odin_state = lambda timeout: True

    test_result = {
        "centre_of_mass": [6, 6, 6],
        "max_voxel": [5, 5, 5],
        "max_count": 123456,
        "n_voxels": 321,
        "total_count": 999999,
        "bounding_box": [[3, 3, 3], [9, 9, 9]],
    }

    @AsyncStatus.wrap
    async def mock_complete(result):
        await fake_composite.zocalo._put_results([result], {"dcid": 0, "dcgid": 0})

    fake_composite.zocalo.trigger = MagicMock(
        side_effect=partial(mock_complete, test_result)
    )  # type: ignore
    fake_composite.zocalo.timeout_s = 3
    set_mock_value(fake_composite.gonio.x.max_velocity, 10)

    set_mock_value(fake_composite.robot.barcode, "BARCODE")

    return fake_composite


def fake_read(obj, initial_positions, _):
    initial_positions[obj] = 0
    yield Msg("null", obj)


def extract_metafile(input_filename, output_filename):
    with gzip.open(input_filename) as metafile_fo:
        with open(output_filename, "wb") as output_fo:
            output_fo.write(metafile_fo.read())


@pytest.fixture
def sim_run_engine():
    logging.getLogger("asyncio").setLevel(logging.DEBUG)
    return RunEngineSimulator()


class DocumentCapturer:
    """A utility which can be subscribed to the RunEngine in place of a callback in order
    to intercept documents and make assertions about their contents"""

    def __init__(self) -> None:
        self.docs_received: list[tuple[str, dict[str, Any]]] = []

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.docs_received.append((args[0], args[1]))

    @staticmethod
    def is_match(
        doc: tuple[str, dict[str, Any]],
        name: str,
        has_fields: Sequence[str] = [],
        matches_fields: dict[str, Any] = {},  # noqa
    ):
        """Returns True if the given document:
        - has the same name
        - contains all the fields in has_fields
        - contains all the fields in matches_fields with the same content"""

        return (
            doc[0] == name
            and all(f in doc[1].keys() for f in has_fields)
            and matches_fields.items() <= doc[1].items()
        )

    @staticmethod
    def get_matches(
        docs: list[tuple[str, dict[str, Any]]],
        name: str,
        has_fields: Sequence[str] = [],
        matches_fields: dict[str, Any] = {},  # noqa
    ):
        """Get all the docs from docs which:
        - have the same name
        - contain all the fields in has_fields
        - contain all the fields in matches_fields with the same content"""
        return list(
            filter(
                partial(
                    DocumentCapturer.is_match,
                    name=name,
                    has_fields=has_fields,
                    matches_fields=matches_fields,
                ),
                docs,
            )
        )

    @staticmethod
    def assert_doc(
        docs: list[tuple[str, dict[str, Any]]],
        name: str,
        has_fields: Sequence[str] = [],
        matches_fields: dict[str, Any] = {},  # noqa
        does_exist: bool = True,
    ):
        """Assert that a matching doc has been received by the sim,
        and returns the first match if it is meant to exist"""
        matches = DocumentCapturer.get_matches(docs, name, has_fields, matches_fields)
        if does_exist:
            assert matches
            return matches[0]
        else:
            assert matches == []

    @staticmethod
    def get_docs_until(
        docs: list[tuple[str, dict[str, Any]]],
        name: str,
        has_fields: Sequence[str] = [],
        matches_fields: dict[str, Any] = {},  # noqa
    ):
        """return all the docs from the list of docs until the first matching one"""
        for i, doc in enumerate(docs):
            if DocumentCapturer.is_match(doc, name, has_fields, matches_fields):
                return docs[: i + 1]
        raise ValueError(f"Doc {name=}, {has_fields=}, {matches_fields=} not found")

    @staticmethod
    def get_docs_from(
        docs: list[tuple[str, dict[str, Any]]],
        name: str,
        has_fields: Sequence[str] = [],
        matches_fields: dict[str, Any] = {},  # noqa
    ):
        """return all the docs from the list of docs after the first matching one"""
        for i, doc in enumerate(docs):
            if DocumentCapturer.is_match(doc, name, has_fields, matches_fields):
                return docs[i:]
        raise ValueError(f"Doc {name=}, {has_fields=}, {matches_fields=} not found")

    @staticmethod
    def assert_events_and_data_in_order(
        docs: list[tuple[str, dict[str, Any]]],
        match_data_keys_list: Sequence[Sequence[str]],
    ):
        for event_data_keys in match_data_keys_list:
            docs = DocumentCapturer.get_docs_from(docs, "event")
            doc = docs.pop(0)[1]["data"]
            assert all(k in doc.keys() for k in event_data_keys), (
                f"One of {event_data_keys=} not in {doc}"
            )


def assert_none_matching(
    messages: list[Msg],
    predicate: Callable[[Msg], bool],
):
    assert not list(filter(predicate, messages))


def pin_tip_edge_data():
    tip_x_px = 130
    tip_y_px = 200
    microns_per_pixel = 2.87  # from zoom levels .xml
    grid_width_px = int(400 / microns_per_pixel)
    target_grid_height_px = 140
    top_edge_data = ([0] * tip_x_px) + (
        [(tip_y_px - target_grid_height_px // 2)] * grid_width_px
    )
    bottom_edge_data = [0] * tip_x_px + [
        (tip_y_px + target_grid_height_px // 2)
    ] * grid_width_px
    top_edge_array = numpy.array(top_edge_data, dtype=numpy.uint32)
    bottom_edge_array = numpy.array(bottom_edge_data, dtype=numpy.uint32)
    return tip_x_px, tip_y_px, top_edge_array, bottom_edge_array


def thin_pin_edges():
    return pin_tip_edge_data()


def fat_pin_edges():
    tip_x_px, tip_y_px, top_edge_array, bottom_edge_array = pin_tip_edge_data()
    bottom_edge_array += 60
    return tip_x_px, tip_y_px, top_edge_array, bottom_edge_array


T = TypeVar("T")


def fake_generator(return_val: T = None) -> MsgGenerator[T]:
    yield from bps.null()
    return return_val


@pytest.fixture
def pin_tip_detection_with_found_pin(ophyd_pin_tip_detection: PinTipDetection):
    @AsyncStatus.wrap
    async def set_good_position():
        x, y, top_edge_array, bottom_edge_array = pin_tip_edge_data()
        set_mock_value(ophyd_pin_tip_detection.triggered_tip, numpy.array([x, y]))
        set_mock_value(ophyd_pin_tip_detection.triggered_top_edge, top_edge_array)  # type: ignore
        set_mock_value(ophyd_pin_tip_detection.triggered_bottom_edge, bottom_edge_array)  # type: ignore

    with patch.object(
        ophyd_pin_tip_detection,
        "trigger",
        side_effect=set_good_position,
    ):
        yield ophyd_pin_tip_detection


# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[Any]):
        if call.excinfo is not None:
            raise call.excinfo.value
        else:
            raise RuntimeError(
                f"{call} has no exception data, an unknown error has occurred"
            )

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[Any]):
        raise excinfo.value


def simulate_xrc_result(
    sim_run_engine: RunEngineSimulator,
    zocalo: ZocaloResults,
    test_results: Sequence[dict],
):
    for k in test_results[0].keys():
        sim_run_engine.add_read_handler_for(
            getattr(zocalo, k), numpy.array([r[k] for r in test_results])
        )


# The remaining code in this conftest is utility for external interaction tests. See https://github.com/DiamondLightSource/mx-bluesky/issues/699 for
# a better organisation of this


def default_raw_gridscan_params(
    tmp_path,
    json_file="tests/test_data/parameter_json_files/test_gridscan_param_defaults.json",
):
    return raw_params_from_file(json_file, tmp_path)


def _dummy_params(tmp_path):
    dummy_params = SpecifiedThreeDGridScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/test_gridscan_param_defaults.json",
            tmp_path,
        )
    )
    return dummy_params


def mx_acquisition_from_conn(mock_ispyb_conn) -> MagicMock:
    return mock_ispyb_conn.return_value.__enter__.return_value.mx_acquisition


def remap_upsert_columns(keys: Sequence[str], values: list):
    return dict(zip(keys, values, strict=False))


class OavGridSnapshotTestEvents:
    test_descriptor_document_oav_snapshot: EventDescriptor = {
        "uid": "b5ba4aec-de49-4970-81a4-b4a847391d34",
        "run_start": "d8bee3ee-f614-4e7a-a516-25d6b9e87ef3",
        "name": DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED,
    }  # type: ignore
    test_event_document_oav_snapshot_xy: Event = {
        "descriptor": "b5ba4aec-de49-4970-81a4-b4a847391d34",
        "time": 1666604299.828203,
        "timestamps": {},
        "seq_num": 1,
        "uid": "29033ecf-e052-43dd-98af-c7cdd62e8174",
        "data": {
            "oav-grid_snapshot-top_left_x": 50,
            "oav-grid_snapshot-top_left_y": 100,
            "oav-grid_snapshot-num_boxes_x": 40,
            "oav-grid_snapshot-num_boxes_y": 20,
            "oav-microns_per_pixel_x": 1.58,
            "oav-microns_per_pixel_y": 1.58,
            "oav-beam_centre_i": 517,
            "oav-beam_centre_j": 350,
            "oav-x_direction": -1,
            "oav-y_direction": -1,
            "oav-z_direction": 1,
            "oav-grid_snapshot-box_width": 0.1 * 1000 / 1.25,  # size in pixels
            "oav-grid_snapshot-last_path_full_overlay": "test_1_y",
            "oav-grid_snapshot-last_path_outer": "test_2_y",
            "oav-grid_snapshot-last_saved_path": "test_3_y",
            "gonio-omega": 0,
            "gonio-chi": 0,
            "gonio-x": 0,
            "gonio-y": 0,
            "gonio-z": 0,
        },
    }
    test_event_document_oav_snapshot_xz: Event = {
        "descriptor": "b5ba4aec-de49-4970-81a4-b4a847391d34",
        "time": 1666604299.828203,
        "timestamps": {},
        "seq_num": 1,
        "uid": "29033ecf-e052-43dd-98af-c7cdd62e8174",
        "data": {
            "oav-grid_snapshot-top_left_x": 50,
            "oav-grid_snapshot-top_left_y": 0,
            "oav-grid_snapshot-num_boxes_x": 40,
            "oav-grid_snapshot-num_boxes_y": 10,
            "oav-grid_snapshot-box_width": 0.1 * 1000 / 1.25,  # size in pixels
            "oav-grid_snapshot-last_path_full_overlay": "test_1_z",
            "oav-grid_snapshot-last_path_outer": "test_2_z",
            "oav-grid_snapshot-last_saved_path": "test_3_z",
            "oav-microns_per_pixel_x": 1.58,
            "oav-microns_per_pixel_y": 1.58,
            "oav-beam_centre_i": 517,
            "oav-beam_centre_j": 350,
            "oav-x_direction": -1,
            "oav-y_direction": -1,
            "oav-z_direction": 1,
            "gonio-omega": -90,
            "gonio-chi": 30,
            "gonio-x": 0,
            "gonio-y": 0,
            "gonio-z": 0,
        },
    }


@pytest.fixture()
def test_event_data(tmp_path):
    return _TestEventData(tmp_path)


_UID_GRIDSCAN_OUTER = "d8bee3ee-f614-4e7a-a516-25d6b9e87ef3"
_UID_GRID_DETECT_AND_DO_GRIDSCAN = "41b82023-c271-449d-9543-260da8d85641"
_UID_ROTATION_MAIN = "2093c941-ded1-42c4-ab74-ea99980fbbfd"
_UID_DO_FGS = "636490db-83da-462c-a537-70e6fe416843"
_UID_ROBOT_UNLOAD = "9920cdec-0f13-442c-b985-3c076eeec61f"

_UID_ROBOT_UNLOAD_DESCRIPTOR = "34551604-5bfb-48e4-9eaa-9508261db03c"


class _TestEventData(OavGridSnapshotTestEvents):
    def __init__(self, tmp_path):
        self._tmp_path = tmp_path

    @property
    def test_grid_detect_and_gridscan_start_document(self) -> RunStart:
        return {  # type: ignore
            "uid": _UID_GRID_DETECT_AND_DO_GRIDSCAN,
            "time": 1666604299.6149616,
            "versions": {"ophyd": "1.6.4.post76+g0895f9f", "bluesky": "1.8.3"},
            "scan_id": 1,
            "plan_type": "generator",
            "subplan_name": PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN,
            "mx_bluesky_parameters": _dummy_params(self._tmp_path).model_dump_json(),
        }

    @property
    def test_grid_detect_and_gridscan_stop_document_with_crystal_exception(
        self,
    ) -> RunStop:
        return {
            "run_start": _UID_GRID_DETECT_AND_DO_GRIDSCAN,
            "time": 1666604299.6149616,
            "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
            "exit_status": "fail",
            "reason": f"{CrystalNotFoundError()}",
        }

    @property
    def test_rotation_start_main_document(self):
        return {
            "uid": _UID_ROTATION_MAIN,
            "subplan_name": PlanNameConstants.ROTATION_MAIN,
            "zocalo_environment": EnvironmentConstants.ZOCALO_ENV,
        }

    @property
    def test_gridscan_outer_start_document(self):
        return {
            "uid": _UID_GRIDSCAN_OUTER,
            "time": 1666604299.6149616,
            "versions": {"ophyd": "1.6.4.post76+g0895f9f", "bluesky": "1.8.3"},
            "scan_id": 1,
            "plan_type": "generator",
            "plan_name": PlanNameConstants.GRIDSCAN_OUTER,
            "subplan_name": PlanNameConstants.GRIDSCAN_OUTER,
            "zocalo_environment": EnvironmentConstants.ZOCALO_ENV,
            "mx_bluesky_parameters": _dummy_params(self._tmp_path).model_dump_json(),
        }

    @property
    def test_robot_unload_start_document(self):
        return {
            "uid": _UID_ROBOT_UNLOAD,
            "subplan_name": PlanNameConstants.ROBOT_UNLOAD,
            "metadata": {"visit": TEST_VISIT, "sample_id": TEST_SAMPLE_ID},
            "activate_callbacks": ["RobotLoadISPyBCallback"],
        }

    @property
    def test_rotation_event_document_during_data_collection(self) -> Event:
        return {
            "descriptor": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
            "time": 2666604299.928203,
            "data": {
                "aperture_scatterguard-aperture-x": 15,
                "aperture_scatterguard-aperture-y": 16,
                "aperture_scatterguard-aperture-z": 2,
                "aperture_scatterguard-scatterguard-x": 18,
                "aperture_scatterguard-scatterguard-y": 19,
                "aperture_scatterguard-selected_aperture": ApertureValue.MEDIUM,
                "aperture_scatterguard-diameter": 50,
                "attenuator-actual_transmission": 0.98,
                "flux-flux_reading": 9.81,
                "dcm-energy_in_keV": 11.105,
                "beamsize-x_um": 50.0,
                "beamsize-y_um": 20.0,
                "eiger_cam_roi_mode": False,
                "eiger-ispyb_detector_id": 78,
            },
            "timestamps": {"det1": 1666604299.8220396, "det2": 1666604299.8235943},
            "seq_num": 1,
            "uid": "2093c941-ded1-42c4-ab74-ea99980fbbfd",
            "filled": {},
        }

    @property
    def test_rotation_stop_main_document(self) -> RunStop:
        return {
            "run_start": _UID_ROTATION_MAIN,
            "time": 1666604300.0310638,
            "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
            "exit_status": "success",
            "reason": "Test succeeded",
            "num_events": {"fake_ispyb_params": 1, "primary": 1},
        }

    @property
    def test_do_fgs_start_document(self) -> RunStart:
        specs = create_dummy_scan_spec()
        return {  # type: ignore
            "uid": _UID_DO_FGS,
            "time": 1666604299.6149616,
            "versions": {"ophyd": "1.6.4.post76+g0895f9f", "bluesky": "1.8.3"},
            "scan_id": 1,
            "plan_type": "generator",
            "subplan_name": PlanNameConstants.DO_FGS,
            "omega_to_scan_spec": {
                GridscanPlane.OMEGA_XY: specs[0],
                GridscanPlane.OMEGA_XZ: specs[1],
            },
        }

    @property
    def test_descriptor_document_oav_rotation_snapshot(self) -> EventDescriptor:
        return {
            "uid": "c7d698ce-6d49-4c56-967e-7d081f964573",
            "name": DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED,
        }  # type: ignore

    @property
    def test_descriptor_document_pre_data_collection(self) -> EventDescriptor:
        return {
            "uid": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
            "name": DocDescriptorNames.HARDWARE_READ_PRE,
        }  # type: ignore

    @property
    def test_descriptor_document_during_data_collection(self) -> EventDescriptor:
        return {
            "uid": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
            "name": DocDescriptorNames.HARDWARE_READ_DURING,
        }  # type: ignore

    @property
    def test_descriptor_document_zocalo_hardware(self) -> EventDescriptor:
        return {
            "uid": "f082901b-7453-4150-8ae5-c5f98bb34406",
            "name": DocDescriptorNames.ZOCALO_HW_READ,
        }  # type: ignore

    @property
    def test_descriptor_document_robot_unload(self) -> EventDescriptor:
        return {
            "uid": _UID_ROBOT_UNLOAD_DESCRIPTOR,
            "name": DocDescriptorNames.ROBOT_UPDATE,
        }  # type: ignore

    @property
    def test_event_document_robot_unload(self) -> Event:
        return {
            "descriptor": _UID_ROBOT_UNLOAD_DESCRIPTOR,
            "time": 1666604299.828203,
            "data": {
                "robot-barcode": "123456",
                "robot-current_pin": 1,
                "robot-current_puck": 2,
                "webcam-last_saved_path": "blah",
                "oav-snapshot-last_saved_path": "blah",
            },
        }  # type: ignore

    @property
    def test_event_document_oav_rotation_snapshot(self) -> Event:
        return {
            "descriptor": "c7d698ce-6d49-4c56-967e-7d081f964573",
            "time": 1666604299.828203,
            "timestamps": {},
            "seq_num": 1,
            "uid": "32d7c25c-c310-4292-ac78-36ce6509be3d",
            "data": {"oav-snapshot-last_saved_path": "snapshot_0"},
        }

    @property
    def test_event_document_pre_data_collection(self) -> Event:
        return {
            "descriptor": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
            "time": 1666604299.828203,
            "data": {
                "s4_slit_gaps-xgap": 0.1234,
                "s4_slit_gaps-ygap": 0.2345,
                "synchrotron-synchrotron_mode": SynchrotronMode.USER,
                "undulator-current_gap": 1.234,
                "gonio-x": 0.158435435,
                "gonio-y": 0.023547354,
                "gonio-z": 0.00345684712,
                "dcm-energy_in_keV": 11.105,
            },
            "timestamps": {"det1": 1666604299.8220396, "det2": 1666604299.8235943},
            "seq_num": 1,
            "uid": "29033ecf-e052-43dd-98af-c7cdd62e8173",
            "filled": {},
        }

    @property
    def test_event_document_during_data_collection(self) -> Event:
        return {
            "descriptor": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
            "time": 2666604299.928203,
            "data": {
                "aperture_scatterguard-aperture-x": 15,
                "aperture_scatterguard-aperture-y": 16,
                "aperture_scatterguard-aperture-z": 2,
                "aperture_scatterguard-scatterguard-x": 18,
                "aperture_scatterguard-scatterguard-y": 19,
                "aperture_scatterguard-selected_aperture": ApertureValue.MEDIUM,
                "aperture_scatterguard-diameter": 50,
                "attenuator-actual_transmission": 1,
                "flux-flux_reading": 10,
                "dcm-energy_in_keV": 11.105,
                "eiger_bit_depth": "16",
                "beamsize-x_um": 50.0,
                "beamsize-y_um": 20.0,
                "eiger_cam_roi_mode": True,
                "eiger-ispyb_detector_id": 78,
            },
            "timestamps": {
                "det1": 1666604299.8220396,
                "det2": 1666604299.8235943,
                "eiger_bit_depth": 1666604299.8220396,
            },
            "seq_num": 1,
            "uid": "29033ecf-e052-43dd-98af-c7cdd62e8174",
            "filled": {},
        }

    @property
    def test_event_document_zocalo_hardware(self) -> Event:
        return {
            "uid": "29033ecf-e052-43dd-98af-c7cdd62e8175",
            "time": 1709654583.9770422,
            "data": {"eiger_odin_file_writer_id": "test_path"},
            "timestamps": {"eiger_odin_file_writer_id": 1666604299.8220396},
            "seq_num": 1,
            "filled": {},
            "descriptor": "f082901b-7453-4150-8ae5-c5f98bb34406",
        }

    @property
    def test_gridscan_outer_stop_document(self) -> RunStop:
        return {
            "run_start": _UID_GRIDSCAN_OUTER,
            "time": 1666604300.0310638,
            "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
            "exit_status": "success",
            "reason": "",
            "num_events": {"fake_ispyb_params": 1, "primary": 1},
        }

    @property
    def test_grid_detect_and_gridscan_stop_document(self) -> RunStop:
        return {
            "run_start": _UID_GRID_DETECT_AND_DO_GRIDSCAN,
            "time": 1666604300.0310638,
            "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
            "exit_status": "success",
            "reason": "",
        }

    @property
    def test_do_fgs_stop_document(self) -> RunStop:
        return {
            "run_start": _UID_DO_FGS,
            "time": 1666604300.0310638,
            "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
            "exit_status": "success",
            "reason": "",
            "num_events": {"fake_ispyb_params": 1, "primary": 1},
        }

    @property
    def test_grid_detect_and_gridscan_failed_stop_document(self) -> RunStop:
        return {
            "run_start": _UID_GRID_DETECT_AND_DO_GRIDSCAN,
            "time": 1666604300.0310638,
            "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
            "exit_status": "fail",
            "reason": "could not connect to devices",
            "num_events": {"fake_ispyb_params": 1, "primary": 1},
        }

    @property
    def test_robot_unload_stop_document(self) -> RunStop:
        return {
            "run_start": _UID_ROBOT_UNLOAD,
            "time": 1666604300.0310638,
            "uid": "863e29d5-0b4e-4a54-bd78-02b23f79309e",
            "exit_status": "success",
            "reason": "",
        }


class TestData(OavGridSnapshotTestEvents):
    DUMMY_TIME_STRING: str = "1970-01-01 00:00:00"

    test_result_large = [
        {
            "centre_of_mass": [1, 2, 3],
            "max_voxel": [1, 2, 3],
            "max_count": 105062,
            "n_voxels": 35,
            "total_count": 2387574,
            "bounding_box": [[2, 2, 2], [8, 8, 7]],
        }
    ]
    test_result_medium = [
        {
            "centre_of_mass": [1, 2, 3],
            "max_voxel": [2, 4, 5],
            "max_count": 50000,
            "n_voxels": 35,
            "total_count": 100000,
            "bounding_box": [[1, 2, 3], [3, 4, 4]],
        }
    ]
    test_result_small = [
        {
            "centre_of_mass": [1, 2, 3],
            "max_voxel": [1, 2, 3],
            "max_count": 1000,
            "n_voxels": 35,
            "total_count": 1000,
            "bounding_box": [[2, 2, 2], [3, 3, 3]],
        }
    ]
    test_result_below_threshold = [
        {
            "centre_of_mass": [2, 3, 4],
            "max_voxel": [2, 3, 4],
            "max_count": 2,
            "n_voxels": 1,
            "total_count": 2,
            "bounding_box": [[1, 2, 3], [2, 3, 4]],
        }
    ]


@pytest.fixture
def dummy_rotation_params(tmp_path):
    dummy_params = RotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_one_multi_rotation_scan_parameters.json",
            tmp_path,
        )
    )
    return dummy_params


@pytest.fixture
def test_rotation_params(tmp_path):
    scan = RotationScan(
        **(
            raw_params_from_file(
                "tests/test_data/parameter_json_files/good_test_one_multi_rotation_scan_parameters.json",
                tmp_path,
            )
        )
    )
    return scan


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class XBPMAndTransmissionWrapperComposite:
    undulator: UndulatorInKeV
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator
    dcm: DCM


@pytest.fixture
def xbpm_and_transmission_wrapper_composite(
    undulator: UndulatorInKeV,
    xbpm_feedback: XBPMFeedback,
    attenuator: BinaryFilterAttenuator,
    dcm: DCM,
) -> XBPMAndTransmissionWrapperComposite:
    return XBPMAndTransmissionWrapperComposite(
        undulator, xbpm_feedback, attenuator, dcm
    )


def assert_images_pixelwise_equal(actual, expected):
    with Image.open(expected) as expected_image:
        expected_bytes = expected_image.tobytes()
        with Image.open(actual) as actual_image:
            actual_bytes = actual_image.tobytes()
            # assert tries to be clever and takes forever if this is inlined and
            # the comparison fails
            bytes_expected_bytes = actual_bytes == expected_bytes
            assert bytes_expected_bytes, (
                f"Actual and expected images differ, {actual} != {expected}"
            )


@pytest.fixture(autouse=True)
def mock_alert_service():
    with patch(
        "mx_bluesky.common.external_interaction.alerting._service._alert_service",
        create=True,
    ) as service:
        yield service


@pytest.fixture()
def patch_beamline_env_variable(monkeypatch):
    monkeypatch.setenv("BEAMLINE", "dev")


@pytest.fixture(autouse=True)
def clear_cache():
    get_config_client.cache_clear()


@pytest.fixture(autouse=True)
def patch_get_hyperion_feature_settings():
    fake_path = "tests/test_data/test_domain_properties"
    with patch(
        "mx_bluesky.hyperion.external_interaction.config_server.GDA_DOMAIN_PROPERTIES_PATH",
        str(fake_path),
    ):
        yield



from __future__ import annotations

from functools import partial

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.beamlines.i04.beamsize import Beamsize
from dodal.devices.beamlines.i04.transfocator import Transfocator
from dodal.devices.common_dcm import DoubleCrystalMonochromator
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    ZebraFastGridScanThreeD,
    set_fast_grid_scan_params,
)
from dodal.devices.flux import Flux
from dodal.devices.mx_phase1.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import UndulatorInKeV
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)
from pydantic import BaseModel

from mx_bluesky.beamlines.i04.external_interaction.config_server import (
    get_i04_feature_settings,
)
from mx_bluesky.common.device_setup_plans.setup_zebra_and_shutter import (
    setup_zebra_for_gridscan,
    tidy_up_zebra_after_gridscan,
)
from mx_bluesky.common.experiment_plans.change_aperture_then_move_plan import (
    get_results_and_move_to_xtal,
)
from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    construct_beamline_specific_fast_gridscan_features,
)
from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    grid_detect_then_xray_centre,
)
from mx_bluesky.common.experiment_plans.inner_plans.xrc_results_utils import (
    zocalo_stage_decorator,
)
from mx_bluesky.common.experiment_plans.oav_snapshot_plan import (
    setup_beamline_for_oav,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
    generate_start_info_from_omega_map,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.parameters.components import get_param_version
from mx_bluesky.common.parameters.constants import (
    EnvironmentConstants,
    OavConstants,
    PlanGroupCheckpointConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.device_composites import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.common.parameters.gridscan import (
    GridCommon,
    SpecifiedThreeDGridScan,
)
from mx_bluesky.common.preprocessors.preprocessors import (
    set_transmission_and_trigger_xbpm_feedback_before_collection_decorator,
)
from mx_bluesky.common.utils.exceptions import CrystalNotFoundError
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.utils import (
    fix_transmission_and_exposure_time_for_current_wavelength,
)
from mx_bluesky.common.utils.xrc_result import XRayCentreEventHandler

DEFAULT_XRC_BEAMSIZE_MICRONS = 20


class I04AutoXrcParams(BaseModel):
    sample_id: int
    file_name: str
    visit: str
    detector_distance_mm: float
    storage_directory: str


def _change_beamsize(
    transfocator: Transfocator, beamsize: float, parameters: GridCommon
):
    """i04 always uses the large aperture and changes beamsize with the transfocator.

    An aperture is needed to reduce scatter but the transfocator is best used for beamsize
    changes as it gives more flux compared to a bigger beam with a small aperture.
    """
    parameters.selected_aperture = ApertureValue.LARGE
    yield from bps.abs_set(
        transfocator, beamsize, group=PlanGroupCheckpointConstants.GRID_READY_FOR_DC
    )


# See https://github.com/DiamondLightSource/blueapi/issues/506 for using device composites
def i04_default_grid_detect_and_xray_centre(
    parameters: I04AutoXrcParams,
    aperture_scatterguard: ApertureScatterguard = inject("aperture_scatterguard"),
    attenuator: BinaryFilterAttenuator = inject("attenuator"),
    backlight: Backlight = inject("backlight"),
    beamstop: Beamstop = inject("beamstop"),
    beamsize: Beamsize = inject("beamsize"),
    dcm: DoubleCrystalMonochromator = inject("dcm"),
    zebra_fast_grid_scan: ZebraFastGridScanThreeD = inject("zebra_fast_grid_scan"),
    flux: Flux = inject("flux"),
    oav: OAV = inject("oav"),
    pin_tip_detection: PinTipDetection = inject("pin_tip_detection"),
    s4_slit_gaps: S4SlitGaps = inject("s4_slit_gaps"),
    undulator: UndulatorInKeV = inject("undulator"),
    xbpm_feedback: XBPMFeedback = inject("xbpm_feedback"),
    zebra: Zebra = inject("zebra"),
    robot: BartRobot = inject("robot"),
    sample_shutter: ZebraShutter = inject("sample_shutter"),
    eiger: EigerDetector = inject("eiger"),
    synchrotron: Synchrotron = inject("synchrotron"),
    zocalo: ZocaloResults = inject("zocalo"),
    smargon: Smargon = inject("gonio"),
    detector_motion: DetectorMotion = inject("detector_motion"),
    transfocator: Transfocator = inject("transfocator"),
    oav_config: str = OavConstants.OAV_CONFIG_JSON,
    udc: bool = False,
) -> MsgGenerator:
    """
    A composite plan which:
    - Uses the OAV to draw a virtual grid over the sample and to take snapshots of the sample
    - Scans through the grid to identify the crystal centre
    - Changes the aperture to match the beam size to the crystal size
    - Moves the sample to the crystal centre of mass

    i04's implementation of this plan is very similar to Hyperion. However, since i04
    isn't running in a continuous Bluesky UDC loop, we take additional steps in beamline
    tidy-up.
    """

    composite = GridDetectThenXRayCentreComposite(
        eiger,
        synchrotron,
        smargon,
        aperture_scatterguard,
        attenuator,
        zocalo,
        backlight,
        beamstop,
        beamsize,
        dcm,
        detector_motion,
        zebra_fast_grid_scan,
        flux,
        oav,
        pin_tip_detection,
        s4_slit_gaps,
        undulator,
        xbpm_feedback,
        zebra,
        robot,
        sample_shutter,
    )
    initial_beamsize = yield from bps.rd(transfocator.current_vertical_size_rbv)

    initial_x = yield from bps.rd(smargon.x.user_readback)
    initial_y = yield from bps.rd(smargon.y.user_readback)
    initial_z = yield from bps.rd(smargon.z.user_readback)

    _current_wavelength_a = yield from bps.rd(composite.dcm.wavelength_in_a)
    grid_common_params = _get_grid_common_params(_current_wavelength_a, parameters)

    def tidy_beamline():
        yield from bps.mv(transfocator, initial_beamsize)

        if not udc:
            yield from get_ready_for_oav_and_close_shutter(
                composite.gonio,
                composite.backlight,
                composite.aperture_scatterguard,
                composite.detector_motion,
            )

    @zocalo_stage_decorator(composite.zocalo)
    @bpp.finalize_decorator(tidy_beamline)
    def _inner_grid_detect_then_xrc():
        # These callbacks let us talk to ISPyB and Nexgen. They aren't included in the common plan because
        # Hyperion handles its callbacks differently to BlueAPI-managed plans, see
        # https://github.com/DiamondLightSource/mx-bluesky/issues/1117
        flyscan_event_handler = XRayCentreEventHandler()
        callbacks = *create_gridscan_callbacks(), flyscan_event_handler

        @bpp.subs_decorator(callbacks)
        @verify_undulator_gap_before_run_decorator(composite)
        @set_transmission_and_trigger_xbpm_feedback_before_collection_decorator(
            composite,
            grid_common_params.transmission_frac,
            PlanNameConstants.GRIDSCAN_OUTER,
        )
        def grid_detect_then_xray_centre_with_callbacks():
            yield from grid_detect_then_xray_centre(
                composite=composite,
                parameters=grid_common_params,
                xrc_params_type=SpecifiedThreeDGridScan,
                construct_beamline_specific=construct_i04_specific_features,
                oav_config=oav_config,
            )

        try:
            yield from grid_detect_then_xray_centre_with_callbacks()
            assert isinstance(
                grid_common_params.specified_grid_params, SpecifiedThreeDGridScan
            ), "Specified grid params couldn't be found after grid detection"
            yield from get_results_and_move_to_xtal(
                composite,
                grid_common_params.specified_grid_params,
                flyscan_event_handler,
            )
        except CrystalNotFoundError:
            yield from bps.mv(
                smargon.x, initial_x, smargon.y, initial_y, smargon.z, initial_z
            )
            raise

    yield from _change_beamsize(
        transfocator, DEFAULT_XRC_BEAMSIZE_MICRONS, grid_common_params
    )
    yield from _inner_grid_detect_then_xrc()


def get_ready_for_oav_and_close_shutter(
    smargon: Smargon,
    backlight: Backlight,
    aperture_scatterguard: ApertureScatterguard,
    detector_motion: DetectorMotion,
):
    yield from bps.wait(PlanGroupCheckpointConstants.GRID_READY_FOR_DC)
    group = "get_ready_for_oav_and_close_shutter"
    LOGGER.info("Non-udc tidy: Setting up beamline for OAV")
    yield from setup_beamline_for_oav(
        smargon, backlight, aperture_scatterguard, group=group
    )
    LOGGER.info("Non-udc tidy: Closing detector shutter")
    yield from bps.abs_set(
        detector_motion.shutter,
        0,
        group=group,
    )
    yield from bps.wait(group)


def create_gridscan_callbacks() -> tuple[
    GridscanNexusFileCallback, GridscanISPyBCallback
]:
    return (
        GridscanNexusFileCallback(param_type=SpecifiedThreeDGridScan),
        GridscanISPyBCallback(
            param_type=GridCommon,
            emit=ZocaloCallback(
                PlanNameConstants.DO_FGS,
                EnvironmentConstants.ZOCALO_ENV,
                generate_start_info_from_omega_map,
            ),
        ),
    )


def construct_i04_specific_features(
    xrc_composite: GridDetectThenXRayCentreComposite,
    xrc_parameters: SpecifiedThreeDGridScan,
) -> BeamlineSpecificFGSFeatures:
    """
    Get all the information needed to do the i04 XRC flyscan.
    """
    signals_to_read_pre_flyscan = [
        xrc_composite.undulator.current_gap,
        xrc_composite.synchrotron.synchrotron_mode,
        xrc_composite.s4_slit_gaps,
        xrc_composite.gonio,
        xrc_composite.dcm.energy_in_keV,
    ]

    signals_to_read_during_collection = [
        xrc_composite.aperture_scatterguard,
        xrc_composite.attenuator.actual_transmission,
        xrc_composite.flux.flux_reading,
        xrc_composite.dcm.energy_in_keV,
        xrc_composite.eiger.bit_depth,
        xrc_composite.beamsize,
        xrc_composite.eiger.cam.roi_mode,
        xrc_composite.eiger.ispyb_detector_id,
    ]

    tidy_plan = partial(
        tidy_up_zebra_after_gridscan,
        xrc_composite.zebra,
        xrc_composite.sample_shutter,
        group="flyscan_zebra_tidy",
        wait=True,
    )
    set_flyscan_params_plan = partial(
        set_fast_grid_scan_params,
        xrc_composite.zebra_fast_grid_scan,
        xrc_parameters.fast_gridscan_params,
    )
    fgs_motors = xrc_composite.zebra_fast_grid_scan
    return construct_beamline_specific_fast_gridscan_features(
        partial(
            setup_zebra_for_gridscan,
        ),
        tidy_plan,
        set_flyscan_params_plan,
        fgs_motors,
        signals_to_read_pre_flyscan,
        signals_to_read_during_collection,  # type: ignore # until https://github.com/DiamondLightSource/mx-bluesky/issues/1076
    )


def _get_grid_common_params(
    _current_wavelength_a: float, parameters: I04AutoXrcParams
) -> GridCommon:
    """Calculate scaled transmission and exposure by comparing current beamline energy to default energy"""
    feature_settings = get_i04_feature_settings()
    _assumed_wavelength_a = feature_settings.ASSUMED_WAVELENGTH_IN_A
    _unscaled_transmission = feature_settings.XRC_UNSCALED_TRANSMISSION_FRAC
    _unscaled_exposure_time_s = feature_settings.XRC_UNSCALED_EXPOSURE_TIME_S
    transmission_frac, exposure_time_s = (
        fix_transmission_and_exposure_time_for_current_wavelength(
            _current_wavelength_a,
            _assumed_wavelength_a,
            _unscaled_transmission,
            _unscaled_exposure_time_s,
        )
    )

    return GridCommon(
        sample_id=parameters.sample_id,
        file_name=parameters.file_name,
        visit=parameters.visit,
        detector_distance_mm=parameters.detector_distance_mm,
        storage_directory=parameters.storage_directory,
        transmission_frac=transmission_frac,
        exposure_time_s=exposure_time_s,
        parameter_model_version=get_param_version(),
    )



import os
import shutil
from pathlib import Path
from unittest.mock import MagicMock

import h5py
import pytest
from bluesky import RunEngine
from bluesky import preprocessors as bpp
from h5py import Dataset, Datatype, File, Group
from numpy import dtype

from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    standard_read_hardware_during_collection,
)
from mx_bluesky.common.parameters.rotation import (
    RotationScan,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST

from .....conftest import extract_metafile, raw_params_from_file

TEST_DATA_DIRECTORY = Path("tests/test_data/nexus_files/rotation")
TEST_EXAMPLE_NEXUS_FILE = Path("ins_8_5.nxs")
TEST_NEXUS_FILENAME = "rotation_scan_test_nexus"
TEST_METAFILE = "ins_8_5_meta.h5.gz"
FAKE_DATAFILE = "../fake_data.h5"

h5item = Group | Dataset | File | Datatype


def get_groups(dataset: h5py.File) -> set:
    e = set()

    def add_layer(s: set, d: h5item):
        if isinstance(d, h5py.Group):
            for k in d:
                s.add(d.name)
                add_layer(s, d[k])

    add_layer(e, dataset)
    return e


def has_equiv_in(item: str, groups: set, exception_table: dict[str, set[str]]):
    if item not in exception_table.keys():
        return False
    # one of the items in exception_table[item] must be in the tested groups
    return exception_table[item] & groups != set()


def test_has_equiv_in():
    test_table = {"a": {"b", "c"}}
    assert not has_equiv_in("a", {"x", "y", "z"}, test_table)
    assert has_equiv_in("a", {"x", "y", "c"}, test_table)


FilesAndgroups = tuple[h5py.File, set[str], h5py.File, set[str]]


@pytest.fixture
def files_and_groups(tmp_path, run_engine: RunEngine, fake_create_rotation_devices):
    tmpdir = tmp_path
    filename, run_number = _generate_fake_nexus(
        TEST_NEXUS_FILENAME, run_engine, fake_create_rotation_devices, tmp_path
    )
    extract_metafile(
        str(TEST_DATA_DIRECTORY / TEST_METAFILE),
        f"{tmpdir}/{filename}_{run_number}_meta.h5",
    )
    extract_metafile(
        str(TEST_DATA_DIRECTORY / TEST_METAFILE),
        f"{tmpdir}/ins_8_5_meta.h5",
    )
    new_hyperion_master = tmpdir / f"{filename}_{run_number}.nxs"
    new_gda_master = tmpdir / TEST_EXAMPLE_NEXUS_FILE
    new_gda_data = [tmpdir / f"ins_8_5_00000{n}.h5" for n in [1, 2, 3, 4]]
    new_hyp_data = [
        tmpdir / f"{filename}_{run_number}_00000{n}.h5" for n in [1, 2, 3, 4]
    ]
    shutil.copy(TEST_DATA_DIRECTORY / TEST_EXAMPLE_NEXUS_FILE, new_gda_master)
    [shutil.copy(TEST_DATA_DIRECTORY / FAKE_DATAFILE, d) for d in new_gda_data]
    [shutil.copy(TEST_DATA_DIRECTORY / FAKE_DATAFILE, d) for d in new_hyp_data]

    with (
        h5py.File(new_gda_master, "r") as example_nexus,
        h5py.File(new_hyperion_master, "r") as hyperion_nexus,
    ):
        yield (
            example_nexus,
            get_groups(example_nexus),
            hyperion_nexus,
            get_groups(hyperion_nexus),
        )


GROUPS_EQUIVALENTS_TABLE = {
    "/entry/instrument/source": {"/entry/source"},
    "/entry/instrument/detector_z": {"/entry/instrument/detector/detector_z"},
    "/entry/instrument/transformations": {"/entry/instrument/detector/transformations"},
}
GROUPS_EXCEPTIONS = {"/entry/instrument/attenuator"}


def test_hyperion_rotation_nexus_groups_against_gda(
    files_and_groups: FilesAndgroups,
):
    _, gda_groups, _, hyperion_groups = files_and_groups
    for item in gda_groups:
        assert (
            item in hyperion_groups
            or has_equiv_in(item, hyperion_groups, GROUPS_EQUIVALENTS_TABLE)
            or item in GROUPS_EXCEPTIONS
        )


DATATYPE_EXCEPTION_TABLE = {
    "/entry/instrument/detector/bit_depth_image": (
        dtype("int64"),
        "gda item bit_depth_image not present",
    ),
    "/entry/instrument/detector/depends_on": (dtype("S48"), dtype("S1024")),
    "/entry/instrument/detector/description": (dtype("S9"), dtype("S1024")),
    "/entry/instrument/detector/detector_readout_time": (
        dtype("int64"),
        "gda item detector_readout_time not present",
    ),
    "/entry/instrument/detector/distance": (
        dtype("<f8"),
        "gda item distance not present",
    ),
    "/entry/instrument/detector/photon_energy": (
        dtype("<f8"),
        "gda item photon_energy not present",
    ),
    "/entry/instrument/detector/sensor_material": (dtype("S2"), dtype("S1024")),
    "/entry/instrument/detector/threshold_energy": (
        dtype("<f8"),
        "gda item threshold_energy not present",
    ),
    "/entry/instrument/detector/type": (dtype("S5"), dtype("S1024")),
    "/entry/instrument/detector/underload_value": (
        dtype("int64"),
        "gda item underload_value not present",
    ),
    "/entry/sample/depends_on": (dtype("S33"), dtype("S1024")),
    "/entry/sample/sample_omega/omega_end": (
        dtype("<f8"),
        "gda item omega_end not present",
    ),
    "/entry/sample/sample_omega/omega_increment_set": (
        dtype("<f8"),
        "gda item omega_increment_set not present",
    ),
    "/entry/instrument/name": (dtype("S20"), dtype("S1024")),
    "/entry/end_time_estimated": (
        dtype("S20"),
        "gda item end_time_estimated not present",
    ),
    "/entry/start_time": (dtype("S10"), dtype("S20")),
    "/entry/data/data": (dtype("uint32"), dtype("int32")),
    "/entry/instrument/detector/detectorSpecific/nimages": (
        dtype("int64"),
        dtype("int32"),
    ),
    "/entry/instrument/detector/detectorSpecific/ntrigger": (
        dtype("int64"),
        "gda item ntrigger not present",
    ),
    "/entry/instrument/detector/detectorSpecific/software_version": (
        dtype("S100"),
        "gda item software_version not present",
    ),
    "/entry/instrument/detector/detectorSpecific/x_pixels": (
        dtype("uint32"),
        "gda item x_pixels not present",
    ),
    "/entry/instrument/detector/detectorSpecific/x_pixels_in_detector": (
        dtype("uint32"),
        dtype("int32"),
    ),
    "/entry/instrument/detector/detectorSpecific/y_pixels": (
        dtype("uint32"),
        "gda item y_pixels not present",
    ),
    "/entry/instrument/detector/detectorSpecific/y_pixels_in_detector": (
        dtype("uint32"),
        dtype("int32"),
    ),
    "/entry/instrument/detector/module/data_origin": (dtype("uint32"), dtype("int32")),
    "/entry/instrument/detector/module/data_size": (dtype("uint32"), dtype("int32")),
    "/entry/instrument/detector/module/data_stride": (dtype("uint32"), dtype("int32")),
}


def mockitem(name, item):
    m = MagicMock()
    m.dtype = f"{name} item {item} not present"
    return m


@pytest.mark.timeout(2)
def test_determine_datatype_differences(
    files_and_groups: FilesAndgroups,
):
    gda_nexus, gda_groups, hyperion_nexus, hyperion_groups = files_and_groups
    diffs = {}
    for item in gda_groups:
        # we checked separately if all expected items should be there
        # but we should probably still check the excepted ones here??
        if item in hyperion_groups:
            hyperion_group = hyperion_nexus[item]
            gda_group = gda_nexus[item]
            print(hyperion_group, gda_group)
            assert isinstance(hyperion_group, Group) and isinstance(gda_group, Group)
            for dset_or_attr in hyperion_group:
                hyperion_item = mockitem("hyperion", dset_or_attr)
                gda_item = mockitem("gda", dset_or_attr)
                try:
                    hyperion_item = hyperion_group[dset_or_attr]
                except KeyError:
                    ...  # should probably correlate this with some key table
                try:
                    gda_item = gda_group[dset_or_attr]
                except KeyError:
                    ...  # should probably correlate this with some key table
                if not isinstance(hyperion_item, Group) and not isinstance(
                    hyperion_item, Datatype
                ):
                    assert not isinstance(gda_item, Group) and not isinstance(
                        gda_item, Datatype
                    )

                    hyperion_dtype = hyperion_item.dtype
                    gda_dtype = gda_item.dtype
                    print(
                        item,
                        dset_or_attr,
                        hyperion_dtype,
                        gda_dtype,
                        hyperion_dtype == gda_dtype,
                    )
                    if not hyperion_dtype == gda_dtype:
                        diffs[item + "/" + str(dset_or_attr)] = (
                            hyperion_dtype,
                            gda_dtype,
                        )
    print(diffs)


@pytest.mark.timeout(2)
def test_hyperion_vs_gda_datatypes(
    files_and_groups: FilesAndgroups,
):
    gda_nexus, gda_groups, hyperion_nexus, hyperion_groups = files_and_groups
    for item in gda_groups:
        # we checked separately if all expected items should be there
        # but we should probably still check the excepted ones here??
        if item in hyperion_groups:
            hyperion_group = hyperion_nexus[item]
            gda_group = gda_nexus[item]
            assert isinstance(hyperion_group, Group) and isinstance(gda_group, Group)
            for dset_or_attr in hyperion_group:
                hyperion_item = mockitem("hyperion", dset_or_attr)
                gda_item = mockitem("gda", dset_or_attr)
                try:
                    hyperion_item = hyperion_group[dset_or_attr]
                except KeyError:
                    ...  # should probably correlate this with some key table
                try:
                    gda_item = gda_group[dset_or_attr]
                except KeyError:
                    ...  # should probably correlate this with some key table
                if not isinstance(hyperion_item, Group) and not isinstance(
                    hyperion_item, Datatype
                ):
                    assert not isinstance(gda_item, Group) and not isinstance(
                        gda_item, Datatype
                    )
                    assert (
                        hyperion_item.dtype == gda_item.dtype
                        or DATATYPE_EXCEPTION_TABLE[item + "/" + str(dset_or_attr)]
                        == (hyperion_item.dtype, gda_item.dtype)
                    )


def _test_params(filename_stub, tmp_path: Path):
    params = RotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_one_multi_rotation_scan_parameters.json",
            tmp_path,
        )
    )
    for scan_params in params.rotation_scans:
        scan_params.x_start_um = 0
        scan_params.y_start_um = 0
        scan_params.z_start_um = 0
        scan_params.scan_width_deg = 360
    params.file_name = filename_stub
    params.demand_energy_ev = 12700
    params.storage_directory = str(tmp_path)
    params.exposure_time_s = 0.004
    return params


def _generate_fake_nexus(
    filename,
    run_engine: RunEngine,
    rotation_scan_composite: RotationScanComposite,
    tmp_path: Path,
):
    params = _test_params(filename, tmp_path)
    run_number = params.detector_params.run_number
    filename_stub, run_number = sim_rotation_scan_to_create_nexus(
        params, rotation_scan_composite, filename, run_engine
    )
    return filename_stub, run_number


def sim_rotation_scan_to_create_nexus(
    test_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    filename_stub,
    run_engine: RunEngine,
):
    run_number = test_params.detector_params.run_number
    nexus_filename = f"{filename_stub}_{run_number}.nxs"

    fake_create_rotation_devices.eiger.bit_depth.sim_put(32)  # type: ignore

    run_engine(
        fake_rotation_scan(
            test_params, RotationNexusFileCallback(), fake_create_rotation_devices
        )
    )

    nexus_path = Path(test_params.storage_directory) / nexus_filename
    assert os.path.isfile(nexus_path)
    return filename_stub, run_number


def fake_rotation_scan(
    parameters: RotationScan,
    subscription: RotationNexusFileCallback,
    rotation_devices: RotationScanComposite,
):
    single_scan_parameters = next(parameters.single_rotation_scans)

    @bpp.subs_decorator(subscription)
    @bpp.set_run_key_decorator("rotation_scan_with_cleanup_and_subs")
    @bpp.run_decorator(  # attach experiment metadata to the start document
        md={
            "subplan_name": CONST.PLAN.ROTATION_OUTER,
            "mx_bluesky_parameters": single_scan_parameters.model_dump_json(),
            "activate_callbacks": "RotationNexusFileCallback",
        }
    )
    def plan():
        yield from standard_read_hardware_during_collection(
            rotation_devices.aperture_scatterguard,
            rotation_devices.attenuator,
            rotation_devices.flux,
            rotation_devices.dcm,
            rotation_devices.eiger,
            rotation_devices.beamsize,
        )

    return plan()



"""
PVs
# Last update Jan 2021 by RLO
# Now with Eiger goodness
"""

requested_transmission = "BL24I-OP-ATTN-01:T2A:SETVAL1"


# Eiger
eiger_filepath = "BL24I-EA-EIGER-01:CAM:FilePath"
eiger_filename = "BL24I-EA-EIGER-01:CAM:FileName"
eiger_od_filepath = "BL24I-EA-EIGER-01:OD:FilePath"
eiger_od_filename = "BL24I-EA-EIGER-01:OD:FileName"
eiger_seq_id = "BL24I-EA-EIGER-01:CAM:SequenceId"
eiger_numimages = "BL24I-EA-EIGER-01:CAM:NumImages"
eiger_od_num_capture = "BL24I-EA-EIGER-01:OD:NumCapture"
eiger_numexpimage = "BL24I-EA-EIGER-01:CAM:NumExposures"
eiger_acquiretime = "BL24I-EA-EIGER-01:CAM:AcquireTime"
eiger_acquireperiod = "BL24I-EA-EIGER-01:CAM:AcquirePeriod"
eiger_imagemode = "BL24I-EA-EIGER-01:CAM:ImageMode"
eiger_triggermode = "BL24I-EA-EIGER-01:CAM:TriggerMode"
eiger_numtriggers = "BL24I-EA-EIGER-01:CAM:NumTriggers"
eiger_manualtrigger = "BL24I-EA-EIGER-01:CAM:ManualTrigger"
eiger_trigger = "BL24I-EA-EIGER-01:CAM:Trigger"
eiger_filewriter = "BL24I-EA-EIGER-01:CAM:FWEnable"
eiger_stream = "BL24I-EA-EIGER-01:CAM:StreamEnable"
eiger_monitor = "BL24I-EA-EIGER-01:CAM:MonitorEnable"
eiger_datasource = "BL24I-EA-EIGER-01:CAM:DataSource"
eiger_statuspoll = "BL24I-EA-EIGER-01:CAM:ReadStatus.SCAN"
eiger_roi_mode = "BL24I-EA-EIGER-01:CAM:ROIMode"
eiger_ff = "BL24I-EA-EIGER-01:CAM:FlatfieldApplied"
eiger_compress = "BL24I-EA-EIGER-01:CAM:FWCompression"
eiger_compresstype = "BL24I-EA-EIGER-01:CAM:CompressionAlgo"
eiger_od_compress = "BL24I-EA-EIGER-01:OD:Compression"
eiger_od_datatype = "BL24I-EA-EIGER-01:OD:DataType"
eiger_bitdepthrbv = "BL24I-EA-EIGER-01:CAM:BitDepthImage_RBV"
eiger_countmode = "BL24I-EA-EIGER-01:CAM:CountingMode"
eiger_autosum = "BL24I-EA-EIGER-01:CAM:AutoSummation"
eiger_hdrdetail = "BL24I-EA-EIGER-01:CAM:StreamHdrDetail"
eiger_hdrappen = "BL24I-EA-EIGER-01:CAM:StreamHdrAppendix"
eiger_od_capture = "BL24I-EA-EIGER-01:OD:Capture"
eiger_acquire = "BL24I-EA-EIGER-01:CAM:Acquire"
eiger_wavelength = "BL24I-EA-EIGER-01:CAM:Wavelength"
eiger_detdist = "BL24I-EA-EIGER-01:CAM:DetDist"
eiger_beamx = "BL24I-EA-EIGER-01:CAM:BeamX"
eiger_beamy = "BL24I-EA-EIGER-01:CAM:BeamY"
eiger_omegaincr = "BL24I-EA-EIGER-01:CAM:OmegaIncr"
eiger_od_filename_rbv = "BL24I-EA-EIGER-01:OD:FP:FileName_RBV"

# BL24I beamline-specific labelled PVs
me14e_filter = "BL24I-EA-CHIP-01:FILTER"  # to check
me14e_filepath = "BL24I-MO-CHIP-01:filePath"
me14e_chip_name = "BL24I-MO-CHIP-01:chipName"
me14e_chipcapacity = "BL24I-MO-CHIP-01:chipCapacity"
me14e_blockcapacity = "BL24I-MO-CHIP-01:blockCapacity"
me14e_exptime = "BL24I-MO-CHIP-01:expTime"
me14e_dcdetdist = "BL24I-MO-CHIP-01:detDistance"

# General Purpose PV. Now use STEP13
ioc13_gp1 = "BL24I-MO-IOC-13:GP1"
ioc13_gp2 = "BL24I-MO-IOC-13:GP2"
ioc13_gp3 = "BL24I-MO-IOC-13:GP3"
ioc13_gp4 = "BL24I-MO-IOC-13:GP4"
ioc13_gp5 = "BL24I-MO-IOC-13:GP5"
ioc13_gp6 = "BL24I-MO-IOC-13:GP6"
ioc13_gp7 = "BL24I-MO-IOC-13:GP7"
ioc13_gp8 = "BL24I-MO-IOC-13:GP8"
ioc13_gp9 = "BL24I-MO-IOC-13:GP9"
ioc13_gp10 = "BL24I-MO-IOC-13:GP10"
ioc13_gp11 = "BL24I-MO-IOC-13:GP11"
ioc13_gp12 = "BL24I-MO-IOC-13:GP12"
ioc13_gp13 = "BL24I-MO-IOC-13:GP13"
ioc13_gp14 = "BL24I-MO-IOC-13:GP14"
ioc13_gp15 = "BL24I-MO-IOC-13:GP15"
ioc13_gp16 = "BL24I-MO-IOC-13:GP16"
ioc13_gp17 = "BL24I-MO-IOC-13:GP17"
ioc13_gp18 = "BL24I-MO-IOC-13:GP18"
ioc13_gp19 = "BL24I-MO-IOC-13:GP19"
ioc13_gp20 = "BL24I-MO-IOC-13:GP20"
ioc13_gp21 = "BL24I-MO-IOC-13:GP21"
ioc13_gp22 = "BL24I-MO-IOC-13:GP22"
ioc13_gp23 = "BL24I-MO-IOC-13:GP23"
ioc13_gp24 = "BL24I-MO-IOC-13:GP24"
ioc13_gp25 = "BL24I-MO-IOC-13:GP25"
ioc13_gp26 = "BL24I-MO-IOC-13:GP26"
ioc13_gp27 = "BL24I-MO-IOC-13:GP27"
ioc13_gp28 = "BL24I-MO-IOC-13:GP28"
ioc13_gp29 = "BL24I-MO-IOC-13:GP29"
ioc13_gp30 = "BL24I-MO-IOC-13:GP30"
ioc13_gp31 = "BL24I-MO-IOC-13:GP31"
ioc13_gp32 = "BL24I-MO-IOC-13:GP32"
ioc13_gp33 = "BL24I-MO-IOC-13:GP33"
ioc13_gp34 = "BL24I-MO-IOC-13:GP34"
ioc13_gp35 = "BL24I-MO-IOC-13:GP35"
ioc13_gp36 = "BL24I-MO-IOC-13:GP36"
ioc13_gp37 = "BL24I-MO-IOC-13:GP37"
ioc13_gp38 = "BL24I-MO-IOC-13:GP38"
ioc13_gp39 = "BL24I-MO-IOC-13:GP39"
ioc13_gp40 = "BL24I-MO-IOC-13:GP40"
ioc13_gp41 = "BL24I-MO-IOC-13:GP41"
ioc13_gp42 = "BL24I-MO-IOC-13:GP42"
ioc13_gp43 = "BL24I-MO-IOC-13:GP43"
ioc13_gp44 = "BL24I-MO-IOC-13:GP44"
ioc13_gp45 = "BL24I-MO-IOC-13:GP45"
ioc13_gp46 = "BL24I-MO-IOC-13:GP46"
ioc13_gp47 = "BL24I-MO-IOC-13:GP47"
ioc13_gp48 = "BL24I-MO-IOC-13:GP48"
ioc13_gp49 = "BL24I-MO-IOC-13:GP49"
ioc13_gp50 = "BL24I-MO-IOC-13:GP50"
ioc13_gp51 = "BL24I-MO-IOC-13:GP51"
ioc13_gp52 = "BL24I-MO-IOC-13:GP52"
ioc13_gp53 = "BL24I-MO-IOC-13:GP53"
ioc13_gp54 = "BL24I-MO-IOC-13:GP54"
ioc13_gp55 = "BL24I-MO-IOC-13:GP55"
ioc13_gp56 = "BL24I-MO-IOC-13:GP56"
ioc13_gp57 = "BL24I-MO-IOC-13:GP57"
ioc13_gp58 = "BL24I-MO-IOC-13:GP58"
ioc13_gp59 = "BL24I-MO-IOC-13:GP59"
ioc13_gp60 = "BL24I-MO-IOC-13:GP60"
ioc13_gp61 = "BL24I-MO-IOC-13:GP61"
ioc13_gp62 = "BL24I-MO-IOC-13:GP62"
ioc13_gp63 = "BL24I-MO-IOC-13:GP63"
ioc13_gp64 = "BL24I-MO-IOC-13:GP64"
ioc13_gp65 = "BL24I-MO-IOC-13:GP65"
ioc13_gp66 = "BL24I-MO-IOC-13:GP66"
ioc13_gp67 = "BL24I-MO-IOC-13:GP67"
ioc13_gp68 = "BL24I-MO-IOC-13:GP68"
ioc13_gp69 = "BL24I-MO-IOC-13:GP69"
ioc13_gp70 = "BL24I-MO-IOC-13:GP70"
ioc13_gp71 = "BL24I-MO-IOC-13:GP71"
ioc13_gp72 = "BL24I-MO-IOC-13:GP72"
ioc13_gp73 = "BL24I-MO-IOC-13:GP73"
ioc13_gp74 = "BL24I-MO-IOC-13:GP74"
ioc13_gp75 = "BL24I-MO-IOC-13:GP75"
ioc13_gp76 = "BL24I-MO-IOC-13:GP76"
ioc13_gp77 = "BL24I-MO-IOC-13:GP77"
ioc13_gp78 = "BL24I-MO-IOC-13:GP78"
ioc13_gp79 = "BL24I-MO-IOC-13:GP79"
ioc13_gp80 = "BL24I-MO-IOC-13:GP80"
ioc13_gp81 = "BL24I-MO-IOC-13:GP81"
ioc13_gp82 = "BL24I-MO-IOC-13:GP82"
ioc13_gp83 = "BL24I-MO-IOC-13:GP83"
ioc13_gp84 = "BL24I-MO-IOC-13:GP84"
ioc13_gp85 = "BL24I-MO-IOC-13:GP85"
ioc13_gp86 = "BL24I-MO-IOC-13:GP86"
ioc13_gp87 = "BL24I-MO-IOC-13:GP87"
ioc13_gp88 = "BL24I-MO-IOC-13:GP88"
ioc13_gp89 = "BL24I-MO-IOC-13:GP89"
ioc13_gp90 = "BL24I-MO-IOC-13:GP90"
ioc13_gp91 = "BL24I-MO-IOC-13:GP91"
ioc13_gp92 = "BL24I-MO-IOC-13:GP92"
ioc13_gp93 = "BL24I-MO-IOC-13:GP93"
ioc13_gp94 = "BL24I-MO-IOC-13:GP94"
ioc13_gp95 = "BL24I-MO-IOC-13:GP95"
ioc13_gp96 = "BL24I-MO-IOC-13:GP96"
ioc13_gp97 = "BL24I-MO-IOC-13:GP97"
ioc13_gp98 = "BL24I-MO-IOC-13:GP98"
ioc13_gp99 = "BL24I-MO-IOC-13:GP99"
ioc13_gp100 = "BL24I-MO-IOC-13:GP100"
ioc13_gp101 = "BL24I-MO-IOC-13:GP101"
ioc13_gp102 = "BL24I-MO-IOC-13:GP102"
ioc13_gp103 = "BL24I-MO-IOC-13:GP103"
ioc13_gp104 = "BL24I-MO-IOC-13:GP104"
ioc13_gp105 = "BL24I-MO-IOC-13:GP105"
ioc13_gp106 = "BL24I-MO-IOC-13:GP106"
ioc13_gp107 = "BL24I-MO-IOC-13:GP107"
ioc13_gp108 = "BL24I-MO-IOC-13:GP108"
ioc13_gp109 = "BL24I-MO-IOC-13:GP109"
ioc13_gp110 = "BL24I-MO-IOC-13:GP110"
ioc13_gp111 = "BL24I-MO-IOC-13:GP111"
ioc13_gp112 = "BL24I-MO-IOC-13:GP112"
ioc13_gp113 = "BL24I-MO-IOC-13:GP113"
ioc13_gp114 = "BL24I-MO-IOC-13:GP114"
ioc13_gp115 = "BL24I-MO-IOC-13:GP115"
ioc13_gp116 = "BL24I-MO-IOC-13:GP116"
ioc13_gp117 = "BL24I-MO-IOC-13:GP117"
ioc13_gp118 = "BL24I-MO-IOC-13:GP118"
ioc13_gp119 = "BL24I-MO-IOC-13:GP119"
ioc13_gp120 = "BL24I-MO-IOC-13:GP120"



from time import sleep
from unittest.mock import ANY, Mock, patch

import pytest
from blueapi.client import BlueapiClient
from blueapi.service.model import TaskRequest
from blueapi.worker import ProgressEvent, StatusView

from mx_bluesky.common.external_interaction.alerting import Metadata
from mx_bluesky.hyperion.supervisor._task_monitor import TaskMonitor

from ....conftest import TEST_VISIT


@pytest.fixture(autouse=True)
def default_beamline(monkeypatch):
    monkeypatch.setenv("BEAMLINE", "i03")


@pytest.fixture
def task_request(external_load_centre_collect_params) -> TaskRequest:
    return TaskRequest(
        name="load_centre_collect",
        params={"parameters": external_load_centre_collect_params},
        instrument_session=TEST_VISIT,
    )


@pytest.fixture
def mock_alerting_service():
    with patch(
        "mx_bluesky.hyperion.supervisor._task_monitor.get_alerting_service"
    ) as mock_get_alerting_service:
        mock_alerting_service = mock_get_alerting_service.return_value
        yield mock_alerting_service


@patch(
    "mx_bluesky.hyperion.supervisor._task_monitor.TaskMonitor.DEFAULT_TIMEOUT_S", 0.25
)
def test_task_monitor_alerts_if_waiting_for_beam(mock_alerting_service, task_request):
    blueapi_client = Mock(spec=BlueapiClient)
    monitor = TaskMonitor(blueapi_client, task_request)
    with monitor:
        monitor.on_blueapi_event(
            ProgressEvent(
                task_id="12345",
                statuses={
                    "12345": StatusView(
                        display_name=TaskMonitor.FEEDBACK_STATUS_NAME,
                        current=0,
                        initial=0,
                        target=1,
                    )
                },
            )
        )
        sleep(0.5)
    blueapi_client.abort.assert_not_called()
    mock_alerting_service.raise_alert.assert_called_once_with(
        "Hyperion is paused waiting for beam on i03.",
        "Hyperion has been paused waiting for beam for 0 minutes.",
        {
            Metadata.SAMPLE_ID: 5461074,
            Metadata.VISIT: "cm31105-4",
            Metadata.CONTAINER: 2,
        },
    )


@patch(
    "mx_bluesky.hyperion.supervisor._task_monitor.TaskMonitor.DEFAULT_TIMEOUT_S", 0.25
)
def test_task_monitor_alerts_and_cancels_request_if_stuck_not_waiting_for_beam(
    mock_alerting_service, task_request
):
    blueapi_client = Mock(spec=BlueapiClient)
    monitor = TaskMonitor(blueapi_client, task_request)
    with monitor:
        monitor.on_blueapi_event(
            ProgressEvent(
                task_id="12345",
                statuses={
                    "12345": StatusView(
                        display_name=TaskMonitor.FEEDBACK_STATUS_NAME,
                        current=0,
                        initial=0,
                        target=1,
                    )
                },
            )
        )
        sleep(0.1)
        monitor.on_blueapi_event(
            ProgressEvent(
                task_id="12345",
                statuses={
                    "12345": StatusView(
                        display_name=TaskMonitor.FEEDBACK_STATUS_NAME,
                        current=1,
                        initial=0,
                        target=1,
                    )
                },
            )
        )
        sleep(0.5)
    blueapi_client.abort.assert_called_with(ANY)
    mock_alerting_service.raise_alert.assert_called_once_with(
        "UDC encountered an error on i03",
        "Hyperion Supervisor detected that BlueAPI was stuck for 0.25 seconds.",
        {
            Metadata.SAMPLE_ID: 5461074,
            Metadata.VISIT: "cm31105-4",
            Metadata.CONTAINER: 2,
        },
    )



import json
from contextlib import nullcontext
from unittest.mock import MagicMock, patch

import pytest
from bluesky.preprocessors import run_decorator, subs_decorator
from ophyd_async.core import init_devices
from ophyd_async.epics.core import epics_signal_rw

from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
    GridscanPlane,
    _smargon_omega_to_xyxz_plane,
)
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.hyperion.parameters.gridscan import GridCommonWithHyperionDetectorParams

from .....conftest import (
    EXPECTED_START_TIME,
    TEST_SAMPLE_ID,
    replace_all_tmp_paths,
)
from .....expeye_helpers import (
    DC_COMMENT_RE,
    DC_RE,
    DCG_RE,
    DCGS_RE,
    DCS_RE,
    GRID_RE,
    POSITION_RE,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
)
from ..callbacks.ispyb.test_gridscan_ispyb_store_3d import (
    TEST_PROPOSAL_REF,
    TEST_VISIT_NUMBER,
)

EXPECTED_DATA_COLLECTION_3D_XY = {
    "comments": "MX-Bluesky: Xray centring 1 -",
    "detectorDistance": 100.0,
    "exposureTime": 0.1,
    "imageDirectory": "{tmp_data}/",
    "imagePrefix": "file_name",
    "imageSuffix": "h5",
    "numberOfPasses": 1,
    "overlap": 0,
    "startImageNumber": 1,
    "xBeam": 150.0,
    "yBeam": 160.0,
    "startTime": EXPECTED_START_TIME,
    "sampleId": 364758,
}

EXPECTED_DATA_COLLECTION_3D_XZ = EXPECTED_DATA_COLLECTION_3D_XY | {
    "comments": "MX-Bluesky: Xray centring 2 -",
}


TEST_GRID_INFO_IDS = (56, 57)

EXPECTED_END_TIME = "2024-02-08 14:04:01"


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
class TestXrayCentreISPyBCallback:
    def test_activity_gated_start_3d(self, mock_ispyb_conn, test_event_data, tmp_path):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(
            test_event_data.test_grid_detect_and_gridscan_start_document
        )  # pyright: ignore
        create_dcg_request = mock_ispyb_conn.calls_for(DCGS_RE)[0].request
        assert (
            mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 2) == TEST_PROPOSAL_REF
        )
        assert (
            int(mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 3))
            == TEST_VISIT_NUMBER
        )
        assert json.loads(create_dcg_request.body) == {
            "experimentType": "Mesh3D",
            "sampleId": TEST_SAMPLE_ID,
        }
        create_dc_requests = [c.request for c in mock_ispyb_conn.calls_for(DCS_RE)]
        assert (
            int(mock_ispyb_conn.match(create_dc_requests[0], DCS_RE, 2))
            == TEST_DATA_COLLECTION_GROUP_ID
        )
        assert (
            int(mock_ispyb_conn.match(create_dc_requests[1], DCS_RE, 2))
            == TEST_DATA_COLLECTION_GROUP_ID
        )
        assert json.loads(create_dc_requests[0].body) == replace_all_tmp_paths(
            EXPECTED_DATA_COLLECTION_3D_XY, tmp_path
        )
        assert json.loads(create_dc_requests[1].body) == replace_all_tmp_paths(
            EXPECTED_DATA_COLLECTION_3D_XZ, tmp_path
        )

        assert len(mock_ispyb_conn.calls_for(POSITION_RE)) == 0
        assert len(mock_ispyb_conn.calls_for(GRID_RE)) == 0

    @patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.StoreInIspyb.update_data_collection_group_table",
    )
    def test_reason_provided_if_crystal_not_found_error(
        self, mock_update_data_collection_group_table, mock_ispyb_conn, test_event_data
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(
            test_event_data.test_grid_detect_and_gridscan_start_document
        )  # pyright: ignore
        callback.activity_gated_stop(
            test_event_data.test_grid_detect_and_gridscan_stop_document_with_crystal_exception
        )
        dcids_to_comment_requests = {
            r.dcid: r for r in mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)
        }
        assert dcids_to_comment_requests[TEST_DATA_COLLECTION_IDS[0]].body == {
            "comments": " DataCollection Unsuccessful reason: Diffraction not found, skipping sample.",
        }
        assert (
            mock_update_data_collection_group_table.call_args_list[0][0][0].comments
            == "Diffraction not found, skipping sample."
        )

    def test_hardware_read_event_3d(self, mock_ispyb_conn, test_event_data):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(
            test_event_data.test_grid_detect_and_gridscan_start_document
        )  # pyright: ignore
        callback.activity_gated_descriptor(
            test_event_data.test_descriptor_document_pre_data_collection
        )
        callback.activity_gated_event(
            test_event_data.test_event_document_pre_data_collection
        )
        assert len(mock_ispyb_conn.calls_for(DCG_RE)) == 1

        expected_upsert = {
            "slitGapHorizontal": 0.1234,
            "slitGapVertical": 0.2345,
            "synchrotronMode": "User",
            "undulatorGap1": 1.234,
            "resolution": 1.1830593331191241,
            "wavelength": 1.11647184541378,
        }
        update_dc_requests = mock_ispyb_conn.dc_calls_for(DC_RE)
        assert update_dc_requests[0].dcid == TEST_DATA_COLLECTION_IDS[0]
        assert update_dc_requests[1].dcid == TEST_DATA_COLLECTION_IDS[1]
        assert update_dc_requests[0].body == expected_upsert
        assert update_dc_requests[1].body == expected_upsert

    def test_flux_read_events_3d(self, mock_ispyb_conn, test_event_data):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(
            test_event_data.test_grid_detect_and_gridscan_start_document
        )  # pyright: ignore
        callback.activity_gated_descriptor(
            test_event_data.test_descriptor_document_pre_data_collection
        )
        callback.activity_gated_event(
            test_event_data.test_event_document_pre_data_collection
        )

        callback.activity_gated_descriptor(
            test_event_data.test_descriptor_document_during_data_collection
        )
        callback.activity_gated_event(
            test_event_data.test_event_document_during_data_collection
        )

        update_dc_requests = mock_ispyb_conn.dc_calls_for(DC_RE)[2:]
        expected_payload = {
            "wavelength": 1.11647184541378,
            "transmission": 100,
            "flux": 10,
            "resolution": 1.1830593331191241,
            "beamSizeAtSampleX": 0.05,
            "beamSizeAtSampleY": 0.02,
            "detectorMode": "ROI",
            "detectorId": 78,
        }
        assert update_dc_requests[0].dcid == TEST_DATA_COLLECTION_IDS[0]
        assert update_dc_requests[1].dcid == TEST_DATA_COLLECTION_IDS[1]
        assert update_dc_requests[0].body == expected_payload
        assert update_dc_requests[1].body == expected_payload
        assert len(mock_ispyb_conn.calls_for(POSITION_RE)) == 0
        assert len(mock_ispyb_conn.calls_for(GRID_RE)) == 0

    @pytest.mark.parametrize(
        "snapshot_events, first_comment",
        [
            [
                [
                    "test_event_document_oav_snapshot_xy",
                    "test_event_document_oav_snapshot_xz",
                ],
                "Diffraction grid scan of 40 by 20 by _.",
            ],
            [
                [
                    "test_event_document_oav_snapshot_xz",
                    "test_event_document_oav_snapshot_xy",
                ],
                "Diffraction grid scan of 40 by _ by 10.",
            ],
        ],
        ids=["xy-then-xz", "xz-then-xy"],
    )
    def test_activity_gated_event_oav_snapshot_triggered(
        self,
        mock_ispyb_conn,
        test_event_data,
        snapshot_events: list[str],
        first_comment: str,
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_start(
            test_event_data.test_grid_detect_and_gridscan_start_document
        )  # pyright: ignore

        callback.activity_gated_descriptor(
            test_event_data.test_descriptor_document_oav_snapshot
        )
        for event in [
            getattr(test_event_data, event_name) for event_name in snapshot_events
        ]:
            callback.activity_gated_event(event)

        ids_to_dc_upsert_requests = {
            r.dcid: r for r in mock_ispyb_conn.dc_calls_for(DC_RE)
        }
        assert ids_to_dc_upsert_requests[TEST_DATA_COLLECTION_IDS[0]].body == {
            "numberOfImages": 40 * 20,
            "xtalSnapshotFullPath1": "test_1_y",
            "xtalSnapshotFullPath2": "test_2_y",
            "xtalSnapshotFullPath3": "test_3_y",
            "axisStart": 0,
            "chiStart": 0,
            "omegaStart": 0,
            "axisEnd": 0,
            "axisRange": 0,
            "dataCollectionNumber": 1,
            "fileTemplate": "file_name_1_master.h5",
        }
        dcids_to_append_comment_reqs = {
            r.dcid: r for r in mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)
        }
        assert dcids_to_append_comment_reqs[TEST_DATA_COLLECTION_IDS[0]].body == {
            "comments": " Diffraction grid scan of 40 by 20 "
            "images in 126.4 um by 126.4 um steps. Top left (px): [50,100], "
            "bottom right (px): [3250,1700].",
        }
        assert ids_to_dc_upsert_requests[TEST_DATA_COLLECTION_IDS[1]].body == {
            "numberOfImages": 40 * 10,
            "xtalSnapshotFullPath1": "test_1_z",
            "xtalSnapshotFullPath2": "test_2_z",
            "xtalSnapshotFullPath3": "test_3_z",
            "axisStart": 90,
            "chiStart": 30,
            "omegaStart": 90,
            "axisEnd": 90,
            "axisRange": 0,
            "dataCollectionNumber": 2,
            "fileTemplate": "file_name_2_master.h5",
        }
        assert dcids_to_append_comment_reqs[TEST_DATA_COLLECTION_IDS[1]].body == {
            "comments": " Diffraction grid scan of 40 by 10 "
            "images in 126.4 um by 126.4 um steps. Top left (px): [50,0], "
            "bottom right (px): [3250,800].",
        }
        ids_to_grid_upsert_requests = {
            rq.dcid: rq for rq in (mock_ispyb_conn.dc_calls_for(GRID_RE))
        }
        assert ids_to_grid_upsert_requests[TEST_DATA_COLLECTION_IDS[0]].body == {
            "dx": 0.1264,
            "dy": 0.1264,
            "stepsX": 40,
            "stepsY": 20,
            "micronsPerPixelX": 1.58,
            "micronsPerPixelY": 1.58,
            "snapshotOffsetXPixel": 50,
            "snapshotOffsetYPixel": 100,
            "orientation": "horizontal",
            "snaked": True,
        }
        assert ids_to_grid_upsert_requests[TEST_DATA_COLLECTION_IDS[1]].body == {
            "dx": 0.1264,
            "dy": 0.1264,
            "stepsX": 40,
            "stepsY": 10,
            "micronsPerPixelX": 1.58,
            "micronsPerPixelY": 1.58,
            "snapshotOffsetXPixel": 50,
            "snapshotOffsetYPixel": 0,
            "orientation": "horizontal",
            "snaked": True,
        }

        update_dcg_requests = [c.request for c in mock_ispyb_conn.calls_for(DCG_RE)]
        assert (
            int(mock_ispyb_conn.match(update_dcg_requests[0], DCG_RE, 2))
            == TEST_DATA_COLLECTION_GROUP_ID
        )
        assert json.loads(update_dcg_requests[0].body)["comments"] == first_comment
        assert (
            int(mock_ispyb_conn.match(update_dcg_requests[1], DCG_RE, 2))
            == TEST_DATA_COLLECTION_GROUP_ID
        )
        assert (
            json.loads(update_dcg_requests[1].body)["comments"]
            == "Diffraction grid scan of 40 by 20 by 10."
        )

    async def test_ispyb_callback_handles_read_hardware_in_run_engine(
        self, run_engine, mock_ispyb_conn, dummy_rotation_data_collection_group_info
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback._handle_ispyb_hardware_read = MagicMock()
        callback._handle_ispyb_transmission_flux_read = MagicMock()
        callback.ispyb = MagicMock()
        callback.params = MagicMock()
        callback.data_collection_group_info = dummy_rotation_data_collection_group_info

        with init_devices(mock=True):
            test_readable = epics_signal_rw(str, "pv")

        @subs_decorator(callback)
        @run_decorator(
            md={
                "activate_callbacks": ["GridscanISPyBCallback"],
            },
        )
        def test_plan():
            yield from read_hardware_plan(
                [test_readable], DocDescriptorNames.HARDWARE_READ_PRE
            )
            yield from read_hardware_plan(
                [test_readable], DocDescriptorNames.HARDWARE_READ_DURING
            )

        run_engine(test_plan())

        callback._handle_ispyb_hardware_read.assert_called_once()
        callback._handle_ispyb_transmission_flux_read.assert_called_once()

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.GridscanISPyBCallback._handle_oav_grid_snapshot_triggered",
    )
    @patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.StoreInIspyb.update_deposition",
    )
    @patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.StoreInIspyb.update_data_collection_group_table",
    )
    def test_given_event_doc_before_start_doc_received_then_exception_raised(
        self,
        mock_update_data_collection_group_table,
        mock_update_deposition,
        mock__handle_oav_grid_snapshot_triggered,
        test_event_data,
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.activity_gated_descriptor(
            test_event_data.test_descriptor_document_oav_snapshot
        )
        callback.ispyb = MagicMock()
        callback.params = MagicMock()
        callback.data_collection_group_info = None
        with pytest.raises(AssertionError) as e:
            callback.activity_gated_event(
                test_event_data.test_event_document_oav_snapshot_xy
            )

        assert "No data collection group info" in str(e.value)

    def test_ispyb_callback_clears_state_after_run_stop(
        self, test_event_data, mock_ispyb_conn
    ):
        callback = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        callback.active = True
        callback.start(test_event_data.test_grid_detect_and_gridscan_start_document)  # type: ignore
        callback.descriptor(test_event_data.test_descriptor_document_oav_snapshot)
        callback.event(test_event_data.test_event_document_oav_snapshot_xy)
        callback.event(test_event_data.test_event_document_oav_snapshot_xz)
        callback.start(test_event_data.test_gridscan_outer_start_document)  # type: ignore
        callback.start(test_event_data.test_do_fgs_start_document)  # type: ignore
        callback.descriptor(
            test_event_data.test_descriptor_document_pre_data_collection
        )  # type: ignore
        callback.event(test_event_data.test_event_document_pre_data_collection)
        callback.descriptor(test_event_data.test_descriptor_document_zocalo_hardware)
        callback.event(test_event_data.test_event_document_zocalo_hardware)
        callback.descriptor(
            test_event_data.test_descriptor_document_during_data_collection  # type: ignore
        )
        assert callback._grid_plane_to_id_map
        callback.stop(test_event_data.test_do_fgs_stop_document)
        callback.stop(test_event_data.test_gridscan_outer_stop_document)  # type: ignore
        callback.stop(test_event_data.test_grid_detect_and_gridscan_stop_document)
        assert not callback._grid_plane_to_id_map


@pytest.mark.parametrize(
    "omega, expected_plane",
    [
        [0, GridscanPlane.OMEGA_XY],
        [180, GridscanPlane.OMEGA_XY],
        [-180, GridscanPlane.OMEGA_XY],
        [540, GridscanPlane.OMEGA_XY],
        [90, GridscanPlane.OMEGA_XZ],
        [-90, GridscanPlane.OMEGA_XZ],
        [270, GridscanPlane.OMEGA_XZ],
        [-270, GridscanPlane.OMEGA_XZ],
        [0.999, GridscanPlane.OMEGA_XY],
        [-0.999, GridscanPlane.OMEGA_XY],
        [1.001, AssertionError],
        [-1.001, AssertionError],
        [91.001, AssertionError],
        [90.999, GridscanPlane.OMEGA_XZ],
        [89.999, GridscanPlane.OMEGA_XZ],
    ],
)
def test_smargon_omega_to_xyxz_plane(omega, expected_plane):
    expects_exception = not (isinstance(expected_plane, GridscanPlane))
    raises_or_not = (
        pytest.raises(expected_plane) if expects_exception else (nullcontext())
    )
    with raises_or_not:
        plane = _smargon_omega_to_xyxz_plane(omega)
        assert expects_exception or plane == expected_plane



from __future__ import annotations

from mx_bluesky.common.external_interaction.ispyb.data_model import DataCollectionInfo
from mx_bluesky.common.parameters.rotation import (
    SingleRotationScan,
)


def populate_data_collection_info_for_rotation(params: SingleRotationScan):
    info = DataCollectionInfo(
        chi_start=params.chi_start_deg,
        omega_start=params.omega_start_deg,
        data_collection_number=params.detector_params.run_number,  # type:ignore # the validator always makes this int
        n_images=params.num_images,
        axis_range=params.rotation_increment_deg,
        axis_start=params.omega_start_deg,
        axis_end=(
            params.omega_start_deg
            + params.scan_width_deg * params.rotation_direction.multiplier
        ),
        kappa_start=params.kappa_start_deg,
    )
    return info
